import logging
import shutil
from collections.abc import Iterable
from datetime import datetime
from enum import StrEnum
from pathlib import Path
from types import ModuleType, TracebackType

import pathvalidate
from rich._null_file import NullFile
from rich.console import Console, ConsoleRenderable, RenderableType
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.padding import Padding
from rich.progress import (
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    RenderableColumn,
    SpinnerColumn,
    Task,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Column
from rich.text import Text
from rich.traceback import Traceback

import atopile
import faebryk
import faebryk.libs
import faebryk.libs.logging
from atopile.errors import UserPythonModuleError, _BaseBaseUserException
from faebryk.libs.logging import FLOG_FMT
from faebryk.libs.util import Advancable, ConfigFlag

from . import console

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

COLOR_LOGS = ConfigFlag("COLOR_LOGS", default=False)
NOW = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
_DEFAULT_FORMATTER = logging.Formatter("%(message)s", datefmt="[%X]")
_SHOW_LOG_FILE_PATH_THRESHOLD = 120


# displayed during LoggingStage
ALERT = logging.INFO + 5
logging.addLevelName(ALERT, "ALERT")


class LogHandler(RichHandler):
    """
    A logging handler that renders output with Rich.

    Suppresses frames from tracebacks conditionally depending on the exception type.
    """

    def __init__(
        self,
        *args,
        console: Console,
        rich_tracebacks: bool = True,
        show_path: bool = False,
        tracebacks_suppress: Iterable[str] | None = ["typer"],
        tracebacks_suppress_map: dict[type[BaseException], Iterable[ModuleType]]
        | None = {UserPythonModuleError: [atopile, faebryk]},
        tracebacks_unwrap: list[type[BaseException]] | None = [UserPythonModuleError],
        hide_traceback_types: tuple[type[BaseException], ...] = (
            _BaseBaseUserException,
        ),
        always_show_traceback_types: tuple[type[BaseException], ...] = (
            UserPythonModuleError,
        ),
        traceback_level: int = logging.ERROR,
        force_terminal: bool = False,
        **kwargs,
    ):
        super().__init__(
            *args,
            console=console,
            rich_tracebacks=rich_tracebacks,
            show_path=show_path,
            **kwargs,
        )
        self.tracebacks_suppress = tracebacks_suppress or []
        self.tracebacks_suppress_map = tracebacks_suppress_map or {}
        self.tracebacks_unwrap = tracebacks_unwrap or []
        self.hide_traceback_types = hide_traceback_types
        self.always_show_traceback_types = always_show_traceback_types
        self.traceback_level = traceback_level
        self._logged_exceptions = set()
        self._is_terminal = force_terminal or console.is_terminal

        self.addFilter(
            lambda record: record.name.startswith("atopile")
            or record.name.startswith("faebryk")
        )

    def _get_suppress(
        self, exc_type: type[BaseException] | None
    ) -> Iterable[str | ModuleType]:
        """
        Get extended list of modules to suppress from tracebacks.
        """
        suppress: set[str | ModuleType] = set(self.tracebacks_suppress)
        if exc_type is not None:
            for _type in self.tracebacks_suppress_map:
                if issubclass(exc_type, _type) or isinstance(exc_type, _type):
                    suppress.update(self.tracebacks_suppress_map[_type])
        return list(suppress)

    def _unwrap_chained_exceptions(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> tuple[type[BaseException] | None, BaseException | None, TracebackType | None]:
        def unwrap(exc: BaseException, exc_type: type[BaseException]) -> BaseException:
            """
            Unwraps an exception chain until we reach an exception
            that is not an instance of _type.
            """
            while isinstance(exc, exc_type) and exc.__cause__ is not None:
                exc = exc.__cause__
            return exc

        if exc_type is not None and exc_value is not None:
            while any(issubclass(exc_type, _type) for _type in self.tracebacks_unwrap):
                exc_value = unwrap(exc_value, exc_type)
                exc_type = type(exc_value)
                exc_traceback = exc_value.__traceback__

        return exc_type, exc_value, exc_traceback

    def _get_traceback(self, record: logging.LogRecord) -> Traceback | None:
        if not record.exc_info:
            return None

        exc_type, exc_value, exc_traceback = record.exc_info

        suppress = self._get_suppress(exc_type)

        hide_traceback = (
            isinstance(exc_value, self.hide_traceback_types)
            and not isinstance(exc_value, self.always_show_traceback_types)
        ) or record.levelno < self.traceback_level

        if isinstance(exc_value, UserPythonModuleError):
            exc_type, exc_value, exc_traceback = self._unwrap_chained_exceptions(
                exc_type, exc_value, exc_traceback
            )

        if hide_traceback or exc_type is None or exc_value is None:
            return None

        return Traceback.from_exception(
            exc_type,
            exc_value,
            exc_traceback,
            width=self.tracebacks_width,
            extra_lines=self.tracebacks_extra_lines,
            theme=self.tracebacks_theme,
            word_wrap=self.tracebacks_word_wrap,
            show_locals=self.tracebacks_show_locals,
            locals_max_length=self.locals_max_length,
            locals_max_string=self.locals_max_string,
            suppress=suppress,
        )

    def _render_message(
        self, record: logging.LogRecord, message: str
    ) -> "ConsoleRenderable":
        """Render message text in to Text.

        Args:
            record (LogRecord): logging Record.
            message (str): String containing log message.

        Returns:
            ConsoleRenderable: Renderable to display log message.
        """
        if not self._is_terminal:
            return Text(message)

        use_markdown = getattr(record, "markdown", False)
        use_markup = getattr(record, "markup", self.markup)
        highlighter = getattr(record, "highlighter", self.highlighter)

        if use_markdown:
            message_text = Markdown(message)
        else:
            message_text = Text.from_markup(message) if use_markup else Text(message)

            if highlighter:
                message_text = highlighter(message_text)

            if self.keywords is None:
                self.keywords = self.KEYWORDS

            if self.keywords:
                message_text.highlight_words(self.keywords, "logging.keyword")

        return message_text

    def render_message(
        self, record: logging.LogRecord, message: str
    ) -> ConsoleRenderable:
        if record.exc_info is not None and isinstance(
            (exc := record.exc_info[1]), ConsoleRenderable
        ):
            # UserExceptions are already renderables
            return exc

        return self._render_message(record, message)

    def _get_hashable(self, record: logging.LogRecord) -> tuple | None:
        if exc_info := getattr(record, "exc_info", None):
            _, exc_value, _ = exc_info
            if exc_value and isinstance(exc_value, _BaseBaseUserException):
                return exc_value.get_frozen()
        return None

    def _prepare_emit(self, record: logging.LogRecord) -> ConsoleRenderable:
        traceback = self._get_traceback(record)

        if self.formatter:
            record.message = record.getMessage()
            formatter = self.formatter
            if hasattr(formatter, "usesTime") and formatter.usesTime():
                record.asctime = formatter.formatTime(record, formatter.datefmt)
            message = formatter.formatMessage(record)
        else:
            message = record.getMessage()

        message_renderable = self.render_message(record, message)

        log_renderable = self.render(
            record=record, traceback=traceback, message_renderable=message_renderable
        )

        return log_renderable

    def emit(self, record: logging.LogRecord) -> None:
        """Invoked by logging."""
        hashable = self._get_hashable(record)
        if hashable and hashable in self._logged_exceptions:
            return None

        log_renderable = self._prepare_emit(record)

        if isinstance(self.console.file, NullFile):
            # Handles pythonw, where stdout/stderr are null, and we return NullFile
            # instance from Console.file. In this case, we still want to make a log
            # record even though we won't be writing anything to a file.
            self.handleError(record)
        else:
            try:
                self.console.print(log_renderable)
            except Exception:
                self.handleError(record)
            finally:
                if hashable:
                    self._logged_exceptions.add(hashable)


class LiveLogHandler(LogHandler):
    def __init__(self, status: "LoggingStage", *args, **kwargs):
        super().__init__(*args, console=status._console, **kwargs)
        self.status = status

    def emit(self, record: logging.LogRecord) -> None:
        hashable = self._get_hashable(record)
        if hashable and hashable in self._logged_exceptions:
            return

        try:
            if record.levelno >= logging.ERROR:
                self.status._error_count += 1
            elif record.levelno >= logging.WARNING:
                self.status._warning_count += 1

            self.status.refresh()

            if record.levelno == ALERT:
                self.status.alert(record.getMessage())

        except Exception:
            self.handleError(record)
        finally:
            if hashable:
                self._logged_exceptions.add(hashable)


class IndentedProgress(Progress):
    def __init__(self, *args, indent: int = 20, **kwargs):
        self.indent = indent
        super().__init__(*args, **kwargs)

    def get_renderable(self):
        return Padding(super().get_renderable(), pad=(0, 0, 0, self.indent))


class ShortTimeElapsedColumn(TimeElapsedColumn):
    """Renders time elapsed."""

    def render(self, task: "Task") -> Text:
        """Show time elapsed."""
        elapsed = task.elapsed or 0
        return Text.from_markup(f"[blue][{elapsed:.1f}s][/blue]")


class StyledMofNCompleteColumn(MofNCompleteColumn):
    def render(self, task: "Task") -> Text:
        return Text.from_markup(
            f"[dim]{task.completed}[/dim]{self.separator}[dim]{task.total}[/dim]"
        )


class SpacerColumn(RenderableColumn):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, renderable=Text(), table_column=Column(ratio=1), **kwargs
        )


class CompletableSpinnerColumn(SpinnerColumn):
    class Status(StrEnum):
        SUCCESS = "[green]✓[/green]"
        FAILURE = "[red]✗[/red]"
        WARNING = "[yellow]⚠[/yellow]"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status = None

    def complete(self, status: Status) -> None:
        self.status = status

    def render(self, task: "Task") -> RenderableType:
        text = (
            Text.from_markup(self.status)
            if self.status is not None
            else self.spinner.render(task.get_time())
        )
        return text


class LoggingStage(Advancable):
    _LOG_LEVELS = {
        logging.DEBUG: "debug",
        logging.INFO: "info",
        logging.WARNING: "warning",
        logging.ERROR: "error",
    }

    def __init__(
        self, name: str, description: str, steps: int | None = None, indent: int = 20
    ):
        self.name = name
        self.description = description
        self.indent = indent
        self.steps = steps
        self._console = console.error_console
        self._warning_count = 0
        self._error_count = 0
        self._info_log_path = None
        self._log_handler = None
        self._file_handlers = []
        self._original_handlers = {}
        self._sanitized_name = pathvalidate.sanitize_filename(self.name)
        self._result = None

        self._progress = IndentedProgress(
            *self._get_columns(),
            console=self._console,
            transient=False,
            auto_refresh=True,
            refresh_per_second=10,
            indent=self.indent,
            expand=True,
        )

    def _get_columns(self) -> tuple[ProgressColumn, ...]:
        show_log_file_path = (
            self._info_log_path and self._console.width >= _SHOW_LOG_FILE_PATH_THRESHOLD
        )

        return tuple(
            col
            for col in (
                CompletableSpinnerColumn(),
                TextColumn("{task.description}"),
                StyledMofNCompleteColumn() if self.steps is not None else None,
                ShortTimeElapsedColumn(),
                SpacerColumn(),
                TextColumn(
                    f"[dim]{self._info_log_path}[/dim]" if show_log_file_path else "",
                    justify="right",
                    table_column=Column(justify="right", overflow="ellipsis"),
                ),
            )
            if col is not None
        )

    def _update_columns(self) -> None:
        self._progress.columns = self._get_columns()

    def alert(self, message: str) -> None:
        message = f"[bold][yellow]![/yellow] {message}[/bold]"

        self._console.print(
            Padding(Text.from_markup(message), pad=(0, 0, 0, self.indent)),
            highlight=True,
        )

    def _generate_description(self) -> str:
        problems = []
        if self._error_count > 0:
            plural_e = "s" if self._error_count > 1 else ""
            problems.append(f"[red]{self._error_count} error{plural_e}[/red]")

        if self._warning_count > 0:
            plural_w = "s" if self._warning_count > 1 else ""
            problems.append(f"[yellow]{self._warning_count} warning{plural_w}[/yellow]")

        problems_text = f" ({', '.join(problems)})" if problems else ""
        return f"{self.description}{problems_text}"

    def refresh(self) -> None:
        self._progress.update(self._task_id, description=self._generate_description())

    def set_total(self, total: int | None) -> None:
        self.steps = total
        self._update_columns()
        self._progress.update(self._task_id, total=total)
        self.refresh()

    def advance(self, advance: int = 1) -> None:
        assert self.steps is not None
        self._progress.advance(self._task_id, advance)

    def __enter__(self) -> "LoggingStage":
        self._setup_logging()
        self._update_columns()
        self._progress.start()
        self._task_id = self._progress.add_task(
            self.description, total=self.steps if self.steps is not None else 1
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._restore_logging()

        if exc_type is not None or self._error_count > 0:
            status = CompletableSpinnerColumn.Status.FAILURE
        elif self._warning_count > 0:
            status = CompletableSpinnerColumn.Status.WARNING
        else:
            status = CompletableSpinnerColumn.Status.SUCCESS

        for column in self._progress.columns:
            if isinstance(column, CompletableSpinnerColumn):
                column.complete(status)

        self.refresh()
        self._progress.stop()

    def _create_log_dir(self) -> Path:
        from atopile.config import config

        base_log_dir = Path(config.project.paths.logs) / "archive" / NOW

        try:
            build_cfg = config.build
            log_dir = base_log_dir / pathvalidate.sanitize_filename(build_cfg.name)
        except RuntimeError:
            log_dir = base_log_dir

        log_dir.mkdir(parents=True, exist_ok=True)

        latest_link = Path(config.project.paths.logs) / "latest"
        if latest_link.exists():
            if latest_link.is_symlink():
                latest_link.unlink()
            else:
                shutil.rmtree(latest_link)
        latest_link.symlink_to(base_log_dir, target_is_directory=True)

        return log_dir

    def _setup_logging(self) -> None:
        root_logger = logging.getLogger()

        self._original_level = root_logger.level
        self._original_handlers = {"root": root_logger.handlers.copy()}

        # log all messages to at least one handler
        root_logger.setLevel(logging.DEBUG)

        self._log_handler = LiveLogHandler(self)
        self._log_handler.setFormatter(_DEFAULT_FORMATTER)
        self._log_handler.setLevel(self._original_level)

        log_dir = self._create_log_dir()

        for handler in root_logger.handlers.copy():
            root_logger.removeHandler(handler)

        root_logger.addHandler(self._log_handler)

        self._file_handlers = []
        self._file_handles = {}

        for level, level_name in self._LOG_LEVELS.items():
            log_file = log_dir / f"{self._sanitized_name}.{level_name}.log"

            self._file_handles[level_name] = log_file.open("w", encoding="utf-8")
            file_console = Console(
                file=self._file_handles[level_name],
                width=500,
                force_terminal=COLOR_LOGS.get(),
            )
            file_handler = LogHandler(
                console=file_console,
                force_terminal=COLOR_LOGS.get(),
                markup=False,
            )
            file_handler.setFormatter(_DEFAULT_FORMATTER)
            file_handler.setLevel(level)
            self._file_handlers.append(file_handler)
            root_logger.addHandler(file_handler)

            if level_name == "info":
                self._info_log_path = log_file.relative_to(Path.cwd())

    def _restore_logging(self) -> None:
        if not self._log_handler and not self._file_handlers:
            return

        root_logger = logging.getLogger()
        root_logger.setLevel(self._original_level)
        if self._log_handler in root_logger.handlers:
            root_logger.removeHandler(self._log_handler)

        for file_handler in self._file_handlers:
            if file_handler in root_logger.handlers:
                root_logger.removeHandler(file_handler)
                file_handler.close()

        for file_handle in self._file_handles.values():
            file_handle.close()

        for handler in root_logger.handlers.copy():
            root_logger.removeHandler(handler)

        for handler in self._original_handlers.get("root", []):
            root_logger.addHandler(handler)

        self._original_handlers = {}
        self._original_level = logging.INFO
        self._log_handler = None
        self._file_handlers = []


handler = LogHandler(console=console.error_console)
handler.setFormatter(_DEFAULT_FORMATTER)

if FLOG_FMT:
    faebryk.libs.logging.setup_basic_logging()
else:
    logging.basicConfig(level=logging.INFO, handlers=[handler])

logger = logging.getLogger(__name__)



"""CLI command definition for `ato build`."""

import logging
from typing import TYPE_CHECKING, Annotated

import typer
from more_itertools import first

from atopile.cli.logging_ import NOW, LoggingStage
from atopile.config import config
from atopile.telemetry import log_to_posthog

if TYPE_CHECKING:
    from faebryk.core.module import Module

logger = logging.getLogger(__name__)


@log_to_posthog("cli:build_end")
def build(
    entry: Annotated[str | None, typer.Argument()] = None,
    selected_builds: Annotated[
        list[str], typer.Option("--build", "-b", envvar="ATO_BUILD")
    ] = [],
    target: Annotated[
        list[str], typer.Option("--target", "-t", envvar="ATO_TARGET")
    ] = [],
    option: Annotated[
        list[str], typer.Option("--option", "-o", envvar="ATO_OPTION")
    ] = [],
    frozen: Annotated[
        bool | None,
        typer.Option(
            help="PCB must be rebuilt without changes. Useful in CI",
            envvar="ATO_FROZEN",
        ),
    ] = None,
    keep_picked_parts: bool | None = None,
    keep_net_names: bool | None = None,
    keep_designators: bool | None = None,
    standalone: bool = False,
    open_layout: Annotated[
        bool | None, typer.Option("--open", envvar="ATO_OPEN_LAYOUT")
    ] = None,
):
    """
    Build the specified --target(s) or the targets specified by the build config.
    Optionally specify a different entrypoint with the argument ENTRY.
    eg. `ato build --target my_target path/to/source.ato:module.path`
    """
    from atopile import buildutil
    from atopile.cli.install import check_missing_deps_or_offer_to_install
    from atopile.config import BuildType
    from faebryk.library import _F as F
    from faebryk.libs.exceptions import accumulate, log_user_errors

    config.apply_options(
        entry=entry,
        selected_builds=selected_builds,
        target=target,
        option=option,
        standalone=standalone,
        frozen=frozen,
        keep_picked_parts=keep_picked_parts,
        keep_net_names=keep_net_names,
        keep_designators=keep_designators,
    )

    check_missing_deps_or_offer_to_install()

    if open_layout is not None:
        config.project.open_layout_on_build = open_layout

    logger.info("Saving logs to %s", config.project.paths.logs / NOW)
    with accumulate() as accumulator:
        for build in config.builds:
            with accumulator.collect(), log_user_errors(logger), build:
                logger.info("Building '%s'", config.build.name)
                with LoggingStage(
                    name=f"init-{config.build.name}",
                    description="Initializing app",
                ):
                    match config.build.build_type:
                        case BuildType.ATO:
                            app = _init_ato_app()
                        case BuildType.PYTHON:
                            app = _init_python_app()
                            app.add(F.is_app_root())
                        case _:
                            raise ValueError(
                                f"Unknown build type: {config.build.build_type}"
                            )

                # TODO: add a way to override the following with custom build machinery
                buildutil.build(app)

    logger.info("Build successful! 🚀")

    if config.should_open_layout_on_build():
        selected_build_names = list(config.selected_builds)
        if len(selected_build_names) == 1:
            build = config.project.builds[first(selected_build_names)]
            try:
                from faebryk.libs.app.pcb import open_pcb

                open_pcb(build.paths.layout)
            except FileNotFoundError:
                pass
            except RuntimeError as e:
                logger.info(
                    f"{e.args[0]}\nReload pcb manually by pressing Ctrl+O; Enter"
                )

        elif len(selected_build_names) > 1:
            logger.warning(
                "`--open` option is only supported when building "
                "a single build. It will be ignored."
            )


def _init_python_app() -> "Module":
    """Initialize a specific .py build."""

    from atopile import errors
    from faebryk.libs.util import import_from_path

    try:
        app_class = import_from_path(
            config.build.entry_file_path, config.build.entry_section
        )
    except FileNotFoundError as e:
        raise errors.UserFileNotFoundError(
            f"Cannot find build entry {config.build.address}"
        ) from e
    except Exception as e:
        raise errors.UserPythonModuleError(
            f"Cannot import build entry {config.build.address}"
        ) from e

    if not isinstance(app_class, type):
        raise errors.UserPythonLoadError(
            f"Build entry {config.build.address} is not a module we can instantiate"
        )

    try:
        app = app_class()
    except Exception as e:
        raise errors.UserPythonConstructionError(
            f"Cannot construct build entry {config.build.address}"
        ) from e

    return app


def _init_ato_app() -> "Module":
    """Initialize a specific .ato build."""

    from atopile import front_end
    from atopile.datatypes import TypeRef
    from faebryk.libs.library import L

    node = front_end.bob.build_file(
        config.build.entry_file_path,
        TypeRef.from_path_str(config.build.entry_section),
    )
    assert isinstance(node, L.Module)
    return node



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class TVS(F.Diode):
    reverse_breakdown_voltage = L.p_field(units=P.V)

    # @L.rt_field
    # def pickable(self):
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.TVS,
    #         {
    #             "forward_voltage": self.forward_voltage,
    #             "reverse_working_voltage": self.reverse_working_voltage,
    #             "reverse_leakage_current": self.reverse_leakage_current,
    #             "max_current": self.max_current,
    #             "reverse_breakdown_voltage": self.reverse_breakdown_voltage,
    #         },
    #     )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

from rich.progress import track

import faebryk.library._F as F
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.font import Font
from faebryk.libs.geometry.basic import get_distributed_points_in_polygon

logger = logging.getLogger(__name__)


class FontLayout(Layout):
    def __init__(
        self,
        font: Font,
        font_size: float,
        text: str,
        density: float,
        bbox: tuple[float, float] | None = None,
        scale_to_fit: bool = False,
    ) -> None:
        """
        Create a layout that distributes nodes in a font

        :param font: The font to use
        :param font_size: The font size to use in points
        :param text: The text to distribute
        :param density: The density of the distribution in nodes/point
        :param bbox: The bounding box to distribute the nodes in
        :param scale_to_fit: Whether to scale the font to fit the bounding box
        """
        super().__init__()

        self.font = font

        logger.info(f"Creating font layout for text: {text}")
        polys = font.string_to_polygons(
            text, font_size, bbox=bbox, scale_to_fit=scale_to_fit
        )

        logger.info(f"Finding points in polygons with density: {density}")
        nodes = []
        for p in track(polys, description="Finding points in polygons"):
            nodes.extend(get_distributed_points_in_polygon(polygon=p, density=density))

        logger.info(f"Creating {len(nodes)} nodes in polygons")

        self.coords = [(n.x, n.y) for n in nodes]

    def get_count(self) -> int:
        """
        Get the number of nodes that fit in the font
        """
        return len(self.coords)

    def apply(self, *nodes_to_distribute: Node) -> None:
        """
        Apply the PCB positions to all nodes that are inside the font
        """
        if len(nodes_to_distribute) != len(self.coords):
            logger.warning(
                f"Number of nodes to distribute ({len(nodes_to_distribute)})"
                " does not match"
                f" the number of coordinates ({len(self.coords)})"
            )

        for coord, node in zip(self.coords, nodes_to_distribute):
            node.add(
                F.has_pcb_position_defined_relative_to_parent(
                    (
                        *coord,
                        0,
                        F.has_pcb_position.layer_type.NONE,
                    )
                )
            )

    def __hash__(self) -> int:
        return hash(id(self))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


class requires_external_usage(Trait.decless()):
    class RequiresExternalUsageNotFulfilled(
        F.implements_design_check.UnfulfilledCheckException
    ):
        def __init__(self, nodes: list[Node]):
            super().__init__(
                "Nodes requiring external usage but not used externally: "
                f"{', '.join(mif.get_full_name() for mif in nodes)}",
                nodes=nodes,
            )

    @property
    def fulfilled(self) -> bool:
        obj = self.get_obj(type=ModuleInterface)
        connected_to = set(obj.connected.get_connected_nodes(types=[type(obj)]))
        parent = obj.get_parent()
        # no shared parent possible
        if parent is None:
            return True
        # TODO: disables checks for floating modules
        if parent[0].get_parent() is None:
            return True
        # no connections
        if not connected_to:
            return False
        parent, _ = parent

        for c in connected_to:
            if parent not in {p for p, _ in c.get_hierarchy()}:
                return True

        return False

    def on_obj_set(self):
        if not isinstance(self.obj, ModuleInterface):
            raise NotImplementedError("Only supported on ModuleInterfaces")

        super().on_obj_set()

    design_check: F.implements_design_check

    @F.implements_design_check.register_post_design_check
    def __check_post_design__(self):
        if not self.fulfilled:
            raise requires_external_usage.RequiresExternalUsageNotFulfilled(
                nodes=[self.obj],
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P
from faebryk.libs.util import assert_once  # noqa: F401

logger = logging.getLogger(__name__)


class TPS2116(F.PowerMux):
    """
    2 to 1 1.6 V to 5.5 V, 2.5-A Low IQ Power Mux with Manual and Priority Switchover
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    mode: F.ElectricLogic
    status: F.ElectricLogic

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    descriptive_properties = L.f_field(F.has_explicit_part.by_mfr)(
        "Texas Instruments", "TPS2116DRLR"
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.ti.com/lit/ds/symlink/tps2116.pdf"
    )

    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power_out.lv,
                "2": self.power_out.hv,
                "3": self.power_in[0].hv,
                "4": self.select.line,
                "5": self.mode.line,
                "6": self.power_in[1].hv,
                "7": self.power_out.hv,
                "8": self.status.line,
            },
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power_in[0].lv: ["GND"],
                self.mode.line: ["MODE"],
                self.select.line: ["PR1"],
                self.status.line: ["ST"],
                self.power_in[0].hv: ["VIN1"],
                self.power_in[1].hv: ["VIN2"],
                self.power_out.hv: ["VOUT"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_module_references(
            self, exclude=[self.power_out, self.power_in[1]]
        )

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        for power in [self.power_in[0], self.power_in[1], self.power_out]:
            power.voltage.constrain_subset(L.Range(1.6 * P.V, 5.5 * P.V))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto
from typing import Callable, Sequence

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class Logic74xx(Module):
    class Family(Enum):
        """
        see https://www.ti.com/lit/sg/sdyu001ab/sdyu001ab.pdf?ts=1692776786223
        """

        UC = auto()
        AUP = auto()
        ALVC = auto()
        AUP1T = auto()
        AVC = auto()
        LV1T = auto()
        LVC = auto()
        AC = auto()
        AHC = auto()
        HC = auto()
        LV_A = auto()
        ALB = auto()
        ALVT = auto()
        GTL = auto()
        GTLP = auto()
        LVT = auto()
        VME = auto()
        ABT = auto()
        ABTE = auto()
        ACT = auto()
        AHCT = auto()
        ALS = auto()
        AS = auto()
        BCT = auto()
        F = auto()
        FB = auto()
        FCT = auto()
        HCT = auto()
        LS = auto()
        LV_AT = auto()
        S = auto()
        TTL = auto()
        CD4000 = auto()

    power: F.ElectricPower
    logic_family = L.p_field(domain=L.Domains.ENUM(Family))

    designator = L.f_field(F.has_designator_prefix)(F.has_designator_prefix.Prefix.U)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __init__(
        self,
        gates_factory: Sequence[Callable[[], F.ElectricLogicGate]],
    ) -> None:
        super().__init__()

        self.gates_factory = gates_factory

    @L.rt_field
    def gates(self):
        return [g() for g in self.gates_factory]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class PJ398SM(Module):
    tip: F.Electrical
    sleeve: F.Electrical
    switch: F.Electrical

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class ISO1540_ReferenceDesign(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    isolator: F.ISO1540

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    def __preinit__(self):
        self.isolator.non_iso.power.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(10 * P.uF, 0.01))
        self.isolator.iso.power.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(10 * P.uF, 0.01))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from pathlib import Path

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.util import once


class is_atomic_part(Module.TraitT.decless()):
    def __init__(
        self,
        manufacturer: str,
        partnumber: str,
        footprint: str,
        symbol: str,
        model: str | None = None,
    ) -> None:
        super().__init__()
        self._manufacturer = manufacturer
        self._partnumber = partnumber
        self._footprint = footprint
        self._symbol = symbol
        self._model = model

    lazy: F.is_lazy

    @property
    @once
    def path(self) -> Path:
        from atopile.front_end import from_dsl

        if (from_dsl_ := self.try_get_trait(from_dsl)) is None:
            raise ValueError(
                "No source context found for module with is_atomic_part trait"
            )

        if from_dsl_.src_file is None:
            raise ValueError(
                "No source file found for module with is_atomic_part trait"
            )

        return from_dsl_.src_file.parent

    @property
    def fp_path(self) -> tuple[Path, str]:
        """
        returns path to footprint and library name
        """
        return self.path / self._footprint, self.path.name

    def on_obj_set(self):
        super().on_obj_set()

        obj = self.get_obj(Module)

        fp_path, fp_lib = self.fp_path
        fp = F.KicadFootprint.from_path(fp_path, lib_name=fp_lib)
        obj.get_trait(F.can_attach_to_footprint).attach(fp)

        # TODO symbol



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Shenzhen_Kinghelm_Elec_KH_BNC75_3511(Module):
    """
    Inner hole BNC connector board-end elbow 3GHz -55℃~+155℃ 75Ω 9.5mm
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    line: F.ElectricSignal
    shield: F.Electrical

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    explicit_part = L.f_field(F.has_explicit_part.by_supplier)("C2837588")
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.line.line: ["OUT"],
                self.shield: ["2", "3", "4"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import Sequence

import faebryk.library._F as F
from atopile.cli.logging_ import ALERT
from atopile.config import BuildTargetConfig, UserConfigurationError
from atopile.config import config as Gcfg
from atopile.errors import UserValueError
from faebryk.core.module import Module
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.libs.ato_part import AtoPart
from faebryk.libs.exceptions import UserResourceException, accumulate
from faebryk.libs.kicad.fileformats_common import C_xyr
from faebryk.libs.kicad.fileformats_latest import (
    C_kicad_footprint_file,
    C_kicad_fp_lib_table_file,
    C_kicad_model_file,
    C_kicad_pcb_file,
)
from faebryk.libs.kicad.fileformats_version import kicad_footprint_file
from faebryk.libs.kicad.paths import GLOBAL_FP_DIR_PATH, GLOBAL_FP_LIB_PATH
from faebryk.libs.picker.lcsc import (
    EasyEDA3DModel,
    EasyEDAAPIResponse,
    EasyEDAFootprint,
    EasyEDAPart,
    EasyEDASymbol,
    PickedPartLCSC,
)
from faebryk.libs.util import (
    KeyErrorNotFound,
    find,
    indented_container,
    md_list,
    once,
    re_in,
    robustly_rm_dir,
    sanitize_filepath_part,
)

logger = logging.getLogger(__name__)


class PartIsNotAutoGenerated(Exception):
    """
    Part exists but is not auto-generated.
    """

    def __init__(self, part: AtoPart):
        super().__init__(f"Part `{part.path}` is not purely auto-generated.")
        self.part = part


class PartLifecycle:
    """

    ```
    EasyEDA API => easyeda2kicad => picker/lcsc.py
        -> build/cache/easyeda/parts/<id>/<id>.json => picker/lcsc.py
        -> build/cache/easyeda/parts/<id>/*.kicad*|.step => picker/lcsc.py
        -> <component_lib>/parts/<mfr_pn>/*.kicad*|.step|.ato => transformer.py
        -> <layout_path>/<build>/<build>.kicad_pcb
    ```

    Notes:
     - `layout_path` determined by config.py (defaults to elec/layout)
     - `component_lib` determined by config.py (defaults to build/kicad/libs)
    """

    class EasyEDA_API:
        DELTA_REFRESH = timedelta(days=1)

        @property
        def _PATH(self) -> Path:
            return Gcfg.project.paths.build / "cache" / "parts" / "easyeda"

        def _get_part_path(self, partno: str) -> Path:
            return self._PATH.joinpath(partno) / f"{partno}.json"

        def _exists(self, partno: str) -> bool:
            return self._get_part_path(partno).exists()

        def shall_refresh(self, partno: str) -> bool:
            if not self._exists(partno):
                return True
            date_queried = self.load(partno).query_time
            return date_queried < datetime.now() - self.DELTA_REFRESH

        def ingest(self, partno: str, data: EasyEDAAPIResponse) -> EasyEDAAPIResponse:
            if self._exists(partno):
                existing = self.load(partno)
                # check whether footprint changed
                if existing.packageDetail.updateTime == data.packageDetail.updateTime:
                    return existing
                # TODO question user
                logger.warning(f"Updating cached {partno}")

            path = self._get_part_path(partno)
            data.dump(path)
            return data

        def load(self, partno: str) -> EasyEDAAPIResponse:
            return EasyEDAAPIResponse.load(self._get_part_path(partno))

    class Easyeda2Kicad:
        @property
        def _PATH(self) -> Path:
            return Gcfg.project.paths.build / "cache" / "parts" / "easyeda"

        def _get_part_path(self, partno: str) -> Path:
            return self._PATH / partno

        def get_fp_path(self, partno: str, footprint_name: str) -> Path:
            # public because needed by lcsc.py
            return self._get_part_path(partno) / f"{footprint_name}.kicad_mod"

        def _get_sym_path(self, partno: str, sym_name: str) -> Path:
            return self._get_part_path(partno) / f"{sym_name}.kicad_sym"

        def get_model_path(self, partno: str, model_name: str) -> Path:
            # public because needed by lcsc.py

            # TODO: Note: There are two names for models
            # 1. filename derived from footprint
            # 2. internal step file name
            # The only way to find out the internal is to download and read the step
            return self._get_part_path(partno) / f"{model_name}.step"

        def shall_refresh_model(self, part: EasyEDAPart) -> bool:
            # no model in api
            if not part.model and not part._pre_model:
                return False

            # TODO: consider timing out?
            if self.get_model_path(part.identifier, part.model_name).exists():
                return False

            return True

        def load_model(self, part: EasyEDAPart) -> EasyEDA3DModel:
            return EasyEDA3DModel.load(
                self.get_model_path(part.identifier, part.model_name)
            )

        def ingest_part(self, part: EasyEDAPart) -> EasyEDAPart:
            footprint = part.footprint
            fp_path = self.get_fp_path(part.identifier, part.footprint.base_name)
            if fp_path.exists():
                existing = EasyEDAFootprint.load(fp_path)
                if diff := existing.compare(footprint):
                    logger.warning(
                        f"Updating cached footprint {part.footprint.base_name}:"
                        f" {indented_container(diff, recursive=True)}"
                    )
                    footprint.dump(fp_path)
            else:
                footprint.dump(fp_path)

            if part.model or part._pre_model:
                model_path = self.get_model_path(part.identifier, part.model_name)
                if part.model:
                    part.model.dump(model_path)
                elif model_path.exists():
                    part.model = EasyEDA3DModel.load(model_path)

            # TODO not sure about name
            sym_name = part.symbol.kicad_symbol.name
            sym_path = self._get_sym_path(part.identifier, sym_name)
            if sym_path.exists():
                existing = EasyEDASymbol.load(sym_path)
                if diff := existing.compare(part.symbol):
                    logger.warning(
                        f"Updating cached symbol {sym_name}:"
                        f" {indented_container(diff, recursive=True)}"
                    )
                    part.symbol.dump(sym_path)
            else:
                part.symbol.dump(sym_path)

            return part

    class Library:
        @property
        def _PATH(self) -> Path:
            return Gcfg.project.paths.parts

        def _get_part_identifier(self, part: EasyEDAPart) -> str:
            sanitized = [sanitize_filepath_part(x) for x in part.mfn_pn]
            manufacturer, partno = sanitized
            if not manufacturer:
                manufacturer = "UNKNOWN"
            assert partno

            return f"{manufacturer}_{partno}"

        def _get_part_path(self, part: EasyEDAPart) -> Path:
            return self._PATH / self._get_part_identifier(part)

        def fp_table(
            self, build: BuildTargetConfig
        ) -> tuple[Path, C_kicad_fp_lib_table_file]:
            fp_table_path = build.paths.fp_lib_table
            if not fp_table_path.exists():
                fp_table = C_kicad_fp_lib_table_file.skeleton()
                # load all parts existing parts into new table
                for part_dir in Gcfg.project.paths.parts.iterdir():
                    part_identifier = part_dir.name
                    if not part_dir.is_dir():
                        continue
                    if not (part_dir / (part_identifier + ".ato")).is_file():
                        continue
                    self.__insert_fp_lib(part_identifier, fp_table_path, fp_table)
            else:
                fp_table = C_kicad_fp_lib_table_file.loads(fp_table_path)

            return fp_table_path, fp_table

        def _insert_fp_lib(self, lib_name: str, fppath: Path | None = None):
            for build in Gcfg.project.builds.values():
                fp_table_path, fp_table = self.fp_table(build)
                self.__insert_fp_lib(lib_name, fp_table_path, fp_table, fppath)

        def __insert_fp_lib(
            self,
            lib_name: str,
            fp_table_path: Path,
            fp_table: C_kicad_fp_lib_table_file,
            fppath: Path | None = None,
        ):
            fppath = fppath or self._PATH / lib_name

            if not fppath.resolve().exists():
                raise FileNotFoundError(f"Footprint does not exist: {fppath.resolve()}")

            fpuri = Gcfg.project.get_relative_to_kicad_project(fppath.resolve())

            if lib_name in fp_table.fp_lib_table.libs:
                if fp_table.fp_lib_table.libs[lib_name].uri != fpuri:
                    # TODO better exception
                    raise Exception(
                        f"Footprint library {lib_name} already exists at different"
                        f" location: {fp_table.fp_lib_table.libs[lib_name].uri} != "
                        f"{fpuri}. Manual ingestion required."
                    )
                return fp_table

            lib = C_kicad_fp_lib_table_file.C_fp_lib_table.C_lib(
                name=lib_name,
                type="KiCad",
                uri=fpuri,
                options="",
                descr=f"atopile: part lib: {lib_name}",
            )
            fp_table.fp_lib_table.libs[lib_name] = lib
            # TODO move somewhere else
            if not getattr(self, "_printed_alert", False):
                logger.log(ALERT, "pcbnew restart required (updated fp-lib-table)")
                self._printed_alert = True

            fp_table.dumps(fp_table_path)
            return fp_table

        def ingest_part(self, part: AtoPart) -> AtoPart:
            from faebryk.library.is_auto_generated import _FileManuallyModified

            if part.path.exists():
                try:
                    existing = AtoPart.load(part.path)
                except _FileManuallyModified as ex:
                    raise UserValueError(
                        f"Part `{part.path}` has been manually modified ({ex}). Please"
                        " revert the changes or remove the is_auto_generated trait."
                    ) from ex
                if not existing.auto_generated:
                    raise PartIsNotAutoGenerated(existing)

                if diff := existing.compare(part):
                    str_diff = str(diff)
                    if len(str_diff) > 1000:
                        str_diff = "Too long to show..."
                    else:
                        str_diff = indented_container(diff, recursive=True)
                    logger.warning(
                        f"Updating library part {part.identifier}: {str_diff}"
                    )
                    robustly_rm_dir(part.path)
                    part.dump()
            else:
                part.dump()

            self._insert_fp_lib(part.path.name)

            return part

        def ingest_part_from_easyeda(self, epart: EasyEDAPart) -> AtoPart:
            out_path = self._get_part_path(epart)
            identifier = self._get_part_identifier(epart)

            # cache
            # we assume that during runtime identical part ids are identical
            # only used for performance atm
            if not hasattr(self, "_cache_ingest_part_from_easyeda"):
                self._cache_ingest_part_from_easyeda = dict[str, AtoPart]()
            if identifier in self._cache_ingest_part_from_easyeda:
                return self._cache_ingest_part_from_easyeda[identifier]

            ato_part = AtoPart(
                identifier=identifier,
                mfn=epart.mfn_pn,
                path=out_path,
                fp=epart.footprint.footprint,
                symbol=epart.symbol.symbol,
                model=C_kicad_model_file.loads(epart.model.step)
                if epart.model
                else None,
                auto_generated=AtoPart.AutoGenerated(
                    source=f"easyeda:{epart.identifier}",
                ),
                datasheet=epart.datasheet_url,
                docstring=epart.description,
                pick_part=PickedPartLCSC(
                    manufacturer=epart.mfn_pn[0],
                    partno=epart.mfn_pn[1],
                    supplier_partno=epart.lcsc_id,
                ),
            )
            out = self.ingest_part(ato_part)

            self._cache_ingest_part_from_easyeda[identifier] = out
            return out

        def get_footprint_from_identifier(
            self, identifier: str, component: Module
        ) -> tuple[Path, C_kicad_footprint_file]:
            # TODO this is old code, avoid reading fp-lib-table

            fp_lib_path = Gcfg.build.paths.fp_lib_table

            class LibNotInTable(Exception):
                def __init__(
                    self, *args: object, lib_id: str, lib_table_path: Path
                ) -> None:
                    super().__init__(*args)
                    self.lib_id = lib_id
                    self.lib_table_path = lib_table_path

            def _find_footprint(
                lib_tables: Sequence[os.PathLike], lib_id: str
            ) -> C_kicad_fp_lib_table_file.C_fp_lib_table.C_lib:
                lib_tables = [Path(lib_table) for lib_table in lib_tables]

                err_accumulator = accumulate(LibNotInTable, FileNotFoundError)

                for lib_table_path in lib_tables:
                    with err_accumulator.collect():
                        lib_table = C_kicad_fp_lib_table_file.loads(lib_table_path)
                        try:
                            return find(
                                lib_table.fp_lib_table.libs.values(),
                                lambda x: x.name == lib_id,
                            )
                        except KeyErrorNotFound as ex:
                            raise LibNotInTable(
                                lib_id=lib_id, lib_table_path=lib_table_path
                            ) from ex

                if ex := err_accumulator.get_exception():
                    raise ex

                raise ValueError("No footprint libraries provided")

            lib_id, fp_name = identifier.split(":")
            try:
                lib = _find_footprint([fp_lib_path, GLOBAL_FP_LIB_PATH], lib_id)
            except* (LibNotInTable, FileNotFoundError):
                from atopile.front_end import from_dsl  # TODO: F.is_from_dsl

                if (
                    (is_atomic_part_ := component.try_get_trait(F.is_atomic_part))
                    is not None
                    and (from_dsl_ := is_atomic_part_.try_get_trait(from_dsl))
                    is not None
                    and from_dsl_.src_file is not None
                ):
                    # insert footprint on first use from dependency package
                    try:
                        self._insert_fp_lib(lib_id, from_dsl_.src_file.parent)
                    except FileNotFoundError:
                        raise

                    lib = _find_footprint([fp_lib_path, GLOBAL_FP_LIB_PATH], lib_id)
                else:
                    raise

            dir_path = Path(
                str(lib.uri)
                .replace("${KIPRJMOD}", str(Gcfg.build.paths.fp_lib_table.parent))
                .replace("${KICAD9_FOOTPRINT_DIR}", str(GLOBAL_FP_DIR_PATH))
            )

            path = dir_path / f"{fp_name}.kicad_mod"
            try:
                return path, kicad_footprint_file(path)
            except FileNotFoundError as ex:
                raise UserResourceException(
                    f"Footprint `{fp_name}` doesn't exist in library `{lib_id}`"
                ) from ex

    class PCB:
        def __init__(self) -> None:
            pass

        def ingest_footprint(
            self,
            transformer: PCB_Transformer,
            component: Module,
            logger: logging.Logger,
            insert_point: C_xyr | None = None,
        ) -> tuple[C_kicad_pcb_file.C_kicad_pcb.C_pcb_footprint, bool]:
            # TODO this is old code taken from PCB_Transformer
            # need to decouple some actions from here
            # e.g insert point is not really at the right place here

            # TODO also implement the ingest logic correctly
            # if the footprint is already on the pcb, we should run our typical
            # resolution flow

            lifecycle = PartLifecycle.singleton()

            f_fp = component.get_trait(F.has_footprint).get_footprint()

            # At this point, all footprints MUST have a KiCAD identifier
            fp_id = f_fp.get_trait(F.has_kicad_footprint).get_kicad_footprint()

            # This is the component which is being stuck on the board
            address = component.get_full_name()

            # All modules MUST have a designator by this point
            ref = component.get_trait(F.has_designator).get_designator()

            ## Update existing footprint
            if pcb_fp_t := f_fp.try_get_trait(
                PCB_Transformer.has_linked_kicad_footprint
            ):
                pcb_fp = pcb_fp_t.get_fp()

                # TODO: this is where I have to implement the footprint override
                if True or fp_id != pcb_fp.name:
                    # Copy the data structure so if we later mutate it we don't
                    # end up w/ those changes everywhere
                    _, lib_fp = lifecycle.library.get_footprint_from_identifier(
                        fp_id, component
                    )
                    if existing_hash_prop := pcb_fp.propertys.get(
                        PCB_Transformer._FP_LIB_HASH
                    ):
                        existing_hash = existing_hash_prop.value
                    else:
                        existing_hash = None

                    # If the hash never existed, or it's changed then update the
                    # footprint
                    if (
                        existing_hash is None
                        or existing_hash
                        != PCB_Transformer._hash_lib_fp(lib_fp.footprint)
                    ):
                        logger.info(
                            f"Updating `{pcb_fp.name}`->`{fp_id}` on"
                            f" `{address}` ({ref})",
                            extra={"markdown": True},
                        )
                        # We need to manually override the name because the
                        # footprint's data could've ultimately come from anywhere
                        lib_fp.footprint.name = fp_id
                        transformer.update_footprint_from_lib(pcb_fp, lib_fp.footprint)
                        transformer.bind_footprint(pcb_fp, component)
                new_fp = False

            ## Add new footprint
            else:
                # Components and footprints MUST have the same linking by this point
                # This should be enforced through attach, and bind_footprint
                assert not component.has_trait(
                    PCB_Transformer.has_linked_kicad_footprint
                )

                logger.info(f"Adding `{fp_id}` as `{address}` ({ref})")
                # Copy the data structure so if we later mutate it we don't
                # end up w/ those changes everywhere
                _, lib_fp = lifecycle.library.get_footprint_from_identifier(
                    fp_id, component
                )
                # We need to manually override the name because the
                # footprint's data could've ultimately come from anywhere
                lib_fp.footprint.name = fp_id
                pcb_fp = transformer.insert_footprint(lib_fp.footprint, insert_point)
                transformer.bind_footprint(pcb_fp, component)
                new_fp = True

            def _get_prop_uuid(name: str) -> str | None:
                if name in pcb_fp.propertys:
                    return pcb_fp.propertys[name].uuid
                return None

            ## Apply propertys, Reference and atopile_address
            property_values = {}

            # Take any descriptive properties defined on the component
            if c_props_t := component.try_get_trait(F.has_descriptive_properties):
                for prop_name, prop_value in c_props_t.get_properties().items():
                    if re_in(
                        prop_name,
                        PCB_Transformer.INCLUDE_DESCRIPTIVE_PROPERTIES_FROM_PCB(),
                    ):
                        property_values[prop_name] = prop_value

            if c_props_t := component.try_get_trait(F.has_datasheet):
                property_values["Datasheet"] = c_props_t.get_datasheet()

            property_values["Reference"] = ref

            if value_t := component.try_get_trait(F.has_simple_value_representation):
                property_values["Value"] = value_t.get_value()
            else:
                property_values["Value"] = ""

            property_values["atopile_address"] = component.get_full_name()

            for prop_name, prop_value in property_values.items():
                ### Get old property value, representing non-existent properties as None
                ### If the property value has changed, update it
                if prop := pcb_fp.propertys.get(prop_name):
                    if prop_value != prop.value:
                        logger.info(
                            f"Updating `{prop_name}`->`{prop_value}` on"
                            f" `{address}` ({ref})",
                            extra={"markdown": True},
                        )
                        pcb_fp.propertys[prop_name].value = prop_value

                ### If it's a new property, add it
                else:
                    logger.info(
                        f"Adding `{prop_name}`=`{prop_value}` to `{address}` ({ref})",
                        extra={"markdown": True},
                    )
                    pcb_fp.propertys[prop_name] = transformer._make_fp_property(
                        property_name=prop_name,
                        layer="User.9",
                        value=prop_value,
                        uuid=_get_prop_uuid(prop_name)
                        or PCB_Transformer.gen_uuid(mark=True),
                    )

            return pcb_fp, new_fp

    def __init__(self):
        self.easyeda_api = self.EasyEDA_API()
        self.easyeda2kicad = self.Easyeda2Kicad()
        self.library = self.Library()
        self.pcb = self.PCB()

        self._delete_deprecated_cache()

    def _delete_deprecated_cache(self):
        cfg_paths = Gcfg.project.paths
        build = Gcfg.project.paths.build

        old_fps = cfg_paths.root / "elec" / "footprints"
        old_ecache = build / "cache" / "easyeda"
        old_kifps = build / "kicad" / "libs" / "footprints" / "lcsc.pretty"
        old_ki3d = build / "kicad" / "libs" / "lcsc.3dshapes"

        if old_fps.is_dir() and (fps := list(old_fps.rglob("*.kicad_mod"))):
            raise UserConfigurationError(
                "The `footprints` directory is deprecated and not empty. "
                "Please transition your manual footprints to atomic parts."
                f"\n{md_list(fps)}"
            )

        for path in [old_ecache, old_kifps, old_ki3d]:
            if path.exists():
                logger.warning(f"Deleting deprecated cache {path}")
                robustly_rm_dir(path)

    @classmethod
    @once
    def singleton(cls) -> "PartLifecycle":
        return cls()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.util import times


class ElectricLogicGate(F.LogicGate):
    @L.rt_field
    def inputs(self):
        return times(self._input_cnt, F.ElectricLogic)

    @L.rt_field
    def outputs(self):
        return times(self._output_cnt, F.ElectricLogic)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""Implementation of tool support over LSP."""

from __future__ import annotations

import contextvars
import copy
import json
import logging
import os
import pathlib
import sys
import traceback
from dataclasses import dataclass, field
from importlib.metadata import version as get_package_version
from pathlib import Path
from typing import Any, Optional, Protocol, Sequence

from atopile import front_end
from atopile.config import _find_project_dir
from atopile.datatypes import TypeRef
from atopile.errors import UserException
from atopile.parse_utils import get_src_info_from_token
from atopile.parser import AtoParser as ap
from faebryk.core.node import Node
from faebryk.libs.exceptions import DowngradedExceptionCollector, iter_leaf_exceptions

# **********************************************************
# Utils for interacting with the atopile front-end
# **********************************************************


def init_atopile_config(working_dir: Path) -> None:
    from atopile.config import config

    config.apply_options(entry=None, working_dir=working_dir)


# **********************************************************
# Update sys.path before importing any bundled libraries.
# **********************************************************
def update_sys_path(path_to_add: str, strategy: str) -> None:
    """Add given path to `sys.path`."""
    if path_to_add not in sys.path and os.path.isdir(path_to_add):
        if strategy == "useBundled":
            sys.path.insert(0, path_to_add)
        elif strategy == "fromEnvironment":
            sys.path.append(path_to_add)


# Ensure that we can import LSP libraries, and other bundled libraries.
update_sys_path(
    os.fspath(pathlib.Path(__file__).parent.parent / "libs"),
    os.getenv("LS_IMPORT_STRATEGY", "useBundled"),
)

# **********************************************************
# Imports needed for the language server goes below this.
# **********************************************************
# pylint: disable=wrong-import-position,import-error

import lsprotocol.types as lsp  # noqa: E402
from pygls import server, uris, workspace  # noqa: E402

import atopile.lsp.lsp_jsonrpc as jsonrpc  # noqa: E402
import atopile.lsp.lsp_utils as utils  # noqa: E402

WORKSPACE_SETTINGS = {}
GLOBAL_SETTINGS = {}
RUNNER = pathlib.Path(__file__).parent / "lsp_runner.py"
DISTRIBUTION_NAME = "atopile"

MAX_WORKERS = 5
# TODO: Update the language server name and version.
LSP_SERVER = server.LanguageServer(
    name=DISTRIBUTION_NAME,
    version=get_package_version(DISTRIBUTION_NAME),
    max_workers=MAX_WORKERS,
    # we don't have incremental parsing yet
    text_document_sync_kind=lsp.TextDocumentSyncKind.Full,
)


# **********************************************************
# Tool specific code goes below this.
# **********************************************************

# Reference:
#  LS Protocol:
#  https://microsoft.github.io/language-server-protocol/specifications/specification-3-16/
#
#  Sample implementations:
#  Pylint: https://github.com/microsoft/vscode-pylint/blob/main/bundled/tool
#  Black: https://github.com/microsoft/vscode-black-formatter/blob/main/bundled/tool
#  isort: https://github.com/microsoft/vscode-isort/blob/main/bundled/tool

# e.g, TOOL_MODULE = "pylint"
TOOL_MODULE = "atopile"

# e.g, TOOL_DISPLAY = "Pylint"
TOOL_DISPLAY = "atopile"

# all scenarios.
TOOL_ARGS = []  # default arguments always passed to your tool.


class URIProtocol(Protocol):
    class TextDocument:
        uri: str

    text_document: TextDocument


def get_file(uri: str) -> Path:
    document = LSP_SERVER.workspace.get_text_document(uri)
    return Path(document.path)


def get_file_contents(uri: str) -> tuple[Path, str]:
    document = LSP_SERVER.workspace.get_text_document(uri)
    file_path = Path(document.path)
    source_text = document.source
    return file_path, source_text


def log(msg: Any):
    print(msg, file=sys.stderr)


# **********************************************************
# Linting features start here
# **********************************************************

GRAPHS: dict[str, dict[TypeRef, Node]] = {}
ACTIVE_BUILD_TARGET: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "ACTIVE_BUILD_TARGET", default=None
)


def _convert_exc_to_diagnostic(
    exc: UserException, severity: lsp.DiagnosticSeverity = lsp.DiagnosticSeverity.Error
) -> tuple[Path | None, lsp.Diagnostic]:
    # default to the start of the file
    start_file_path = None
    start_line, start_col = 0, 0
    stop_line, stop_col = 0, 0

    if exc.origin_start is not None:
        start_file_path, start_line, start_col = get_src_info_from_token(
            exc.origin_start
        )

        if exc.origin_stop is not None:
            stop_line, stop_col = (
                exc.origin_stop.line,
                exc.origin_stop.column + len(exc.origin_stop.text),
            )
        else:
            # just extend to the next line
            stop_line, stop_col = start_line + 1, 0

    # convert from 1-indexed (ANTLR) to 0-indexed (LSP)
    start_line = max(start_line - 1, 0)
    stop_line = max(stop_line - 1, 0)

    return Path(start_file_path) if start_file_path else None, lsp.Diagnostic(
        range=lsp.Range(
            start=lsp.Position(line=start_line, character=start_col),
            end=lsp.Position(line=stop_line, character=stop_col),
        ),
        message=exc.message,
        severity=severity,
        code=exc.code,
        source=TOOL_DISPLAY,
        # TODO: tags
    )


def _paths_are_equivalent(path1: Path, path2: Path) -> bool:
    return path1.resolve() == path2.resolve()


def _get_diagnostics(uri: str, identifier: str | None = None) -> list[lsp.Diagnostic]:
    """
    Get static diagnostics for a given URI and identifier.

    TODO: caching
    FIXME: combine with _build_document
    """
    file_path, source_text = get_file_contents(uri)
    exc_diagnostics = []

    with DowngradedExceptionCollector(UserException) as collector:
        try:
            front_end.bob.try_build_all_from_text(source_text, file_path)
        except* UserException as e:
            exc_diagnostics = [
                _convert_exc_to_diagnostic(error) for error in iter_leaf_exceptions(e)
            ]

        warning_diagnostics = [
            _convert_exc_to_diagnostic(error, severity=lsp.DiagnosticSeverity.Warning)
            for error, severity in collector
            if severity == logging.WARNING
        ]

    document_file_path = get_file(uri)

    return [
        diag
        for diag_file_path, diag in exc_diagnostics + warning_diagnostics
        if diag_file_path is None
        or _paths_are_equivalent(document_file_path, diag_file_path)
    ]


def _build_document(uri: str, text: str) -> None:
    init_atopile_config(get_file(uri).parent)

    context = front_end.bob.index_text(text, get_file(uri))

    # TOOD: do something smarter here (only distinct trees?)
    GRAPHS.setdefault(uri, {})
    for ref, ctx in context.refs.items():
        match ctx:
            case ap.AtoParser.BlockdefContext():
                try:
                    # try the single-node version first, in case that's all we can build
                    GRAPHS[uri][TypeRef.from_one("__" + str(ref))] = (
                        front_end.bob.build_node(text, Path(uri), ref)
                    )

                    front_end.bob.reset()

                    GRAPHS[uri][ref] = front_end.bob.build_text(text, Path(uri), ref)
                except* UserException as excs:
                    msg = f"Error(s) building {uri}:{ref}:\n"
                    for exc in iter_leaf_exceptions(excs):
                        msg += f"  {exc.message}\n"
                    log_error(msg)
                except* Exception:
                    import traceback

                    log_error(f"Error building {uri}:{ref}:\n{traceback.format_exc()}")
                finally:
                    front_end.bob.reset()

            case _:  # Node or ImportPlaceholder
                try:
                    GRAPHS[uri][TypeRef.from_one(name="__import__" + str(ref))] = (
                        front_end.bob.build_node(text, Path(uri), ref)
                    )
                except Exception:
                    import traceback

                    log_error(f"Error building {uri}:{ref}:\n{traceback.format_exc()}")
                finally:
                    front_end.bob.reset()


@dataclass
class DidChangeBuildTargetParams:
    buildTarget: str | None = field(default=None)


# TODO: implement something useful
@LSP_SERVER.feature("atopile/didChangeBuildTarget")
def on_did_change_build_target(params: DidChangeBuildTargetParams) -> None: ...


@LSP_SERVER.feature(
    lsp.TEXT_DOCUMENT_DIAGNOSTIC,
    lsp.DiagnosticOptions(
        identifier=TOOL_DISPLAY,
        inter_file_dependencies=True,
        workspace_diagnostics=False,
    ),
)
def on_document_diagnostic(params: lsp.DocumentDiagnosticParams) -> None:
    """Handle document diagnostic request."""
    LSP_SERVER.publish_diagnostics(
        params.text_document.uri, _get_diagnostics(params.text_document.uri)
    )


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def on_document_did_open(params: lsp.DidOpenTextDocumentParams) -> None:
    """Handle document open request."""
    _build_document(params.text_document.uri, params.text_document.text)
    LSP_SERVER.publish_diagnostics(
        params.text_document.uri, _get_diagnostics(params.text_document.uri)
    )


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def on_document_did_change(params: lsp.DidChangeTextDocumentParams) -> None:
    """Handle document change request."""
    _build_document(
        params.text_document.uri,
        LSP_SERVER.workspace.get_text_document(params.text_document.uri).source,
    )
    # TODO: debounce
    LSP_SERVER.publish_diagnostics(
        params.text_document.uri, _get_diagnostics(params.text_document.uri)
    )


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_SAVE)
def on_document_did_save(params: lsp.DidSaveTextDocumentParams) -> None:
    """Handle document save request."""
    LSP_SERVER.publish_diagnostics(
        params.text_document.uri, _get_diagnostics(params.text_document.uri)
    )


def _span_to_lsp_range(span: front_end.Span) -> lsp.Range:
    return lsp.Range(
        start=lsp.Position(line=span.start.line - 1, character=span.start.col),
        end=lsp.Position(line=span.end.line - 1, character=span.end.col),
    )


def _query_params(params: lsp.HoverParams | lsp.DefinitionParams) -> dict[str, Any]:
    return {
        "file_path": f"file:{get_file(params.text_document.uri)}",
        "line": params.position.line + 1,  # 0-indexed -> 1-indexed
        "col": params.position.character,
    }


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_HOVER)
def on_document_hover(params: lsp.HoverParams) -> lsp.Hover | None:
    """Handle document hover request."""
    for root in GRAPHS.get(params.text_document.uri, {}).values():
        for _, trait in root.iter_children_with_trait(front_end.from_dsl):
            if (span := trait.query_references(**_query_params(params))) is not None:
                return lsp.Hover(
                    contents=lsp.MarkupContent(
                        kind=lsp.MarkupKind.Markdown, value=trait.hover_text
                    ),
                    range=_span_to_lsp_range(span),
                )


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DEFINITION)
def on_document_definition(params: lsp.DefinitionParams) -> lsp.LocationLink | None:
    """Handle document definition request."""
    for root in GRAPHS.get(params.text_document.uri, {}).values():
        for _, trait in root.iter_children_with_trait(front_end.from_dsl):
            if (spans := trait.query_definition(**_query_params(params))) is not None:
                origin_span, target_span, target_selection_span = spans
                return lsp.LocationLink(
                    target_uri=target_span.start.file,
                    target_range=_span_to_lsp_range(target_span),
                    target_selection_range=_span_to_lsp_range(target_selection_span),
                    origin_selection_range=_span_to_lsp_range(origin_span),
                )


# TODO: if you want to handle setting specific severity for your linter
# in a user configurable way, then look at look at how it is implemented
# for `pylint` extension from our team.
# Pylint: https://github.com/microsoft/vscode-pylint
# Follow the flow of severity from the settings in package.json to the server.
def _get_severity(*_codes: list[str]) -> lsp.DiagnosticSeverity:
    # TODO: All reported issues from linter are treated as warning.
    # change it as appropriate for your linter.
    return lsp.DiagnosticSeverity.Error


# **********************************************************
# Required Language Server Initialization and Exit handlers.
# **********************************************************
@LSP_SERVER.feature(lsp.INITIALIZE)
def initialize(params: lsp.InitializeParams) -> None:
    """LSP handler for initialize request."""
    log_to_output(f"CWD: {os.getcwd()}")

    paths = "\r\n   ".join(sys.path)
    log_to_output(f"sys.path used to run:\r\n   {paths}")

    if params.initialization_options is not None:
        GLOBAL_SETTINGS.update(
            **params.initialization_options.get("globalSettings", {})
        )

        settings = params.initialization_options["settings"]
        _update_workspace_settings(settings)
        log_to_output(
            f"Settings used to run:{json.dumps(settings, indent=4, ensure_ascii=False)}"
        )

    log_to_output(
        f"Global settings:{json.dumps(GLOBAL_SETTINGS, indent=4, ensure_ascii=False)}"
    )

    log_to_output(
        f"Workspace settings:"
        f"{json.dumps(WORKSPACE_SETTINGS, indent=4, ensure_ascii=False)}"
    )

    workspace_dir = Path(WORKSPACE_SETTINGS.get("workspaceFS", os.getcwd()))
    project_dir = _find_project_dir(start=workspace_dir)
    working_dir = project_dir or workspace_dir

    log_to_output(f"Initializing atopile config for `{working_dir}`")
    init_atopile_config(working_dir)


@LSP_SERVER.feature(lsp.EXIT)
def on_exit(_params: Optional[Any] = None) -> None:
    """Handle clean up on exit."""
    jsonrpc.shutdown_json_rpc()


@LSP_SERVER.feature(lsp.SHUTDOWN)
def on_shutdown(_params: Optional[Any] = None) -> None:
    """Handle clean up on shutdown."""
    jsonrpc.shutdown_json_rpc()


def _get_global_defaults():
    return {
        "path": GLOBAL_SETTINGS.get("path", []),
        "interpreter": GLOBAL_SETTINGS.get("interpreter", [sys.executable]),
        "args": GLOBAL_SETTINGS.get("args", []),
        "importStrategy": GLOBAL_SETTINGS.get("importStrategy", "useBundled"),
        "showNotifications": GLOBAL_SETTINGS.get("showNotifications", "off"),
    }


def _update_workspace_settings(settings):
    if not settings:
        key = os.getcwd()
        WORKSPACE_SETTINGS[key] = {
            "cwd": key,
            "workspaceFS": key,
            "workspace": uris.from_fs_path(key),
            **_get_global_defaults(),
        }
        return

    for setting in settings:
        key = uris.to_fs_path(setting["workspace"])
        WORKSPACE_SETTINGS[key] = {
            "cwd": key,
            **setting,
            "workspaceFS": key,
        }


def _get_settings_by_path(file_path: pathlib.Path):
    workspaces = {s["workspaceFS"] for s in WORKSPACE_SETTINGS.values()}

    while file_path != file_path.parent:
        str_file_path = str(file_path)
        if str_file_path in workspaces:
            return WORKSPACE_SETTINGS[str_file_path]
        file_path = file_path.parent

    setting_values = list(WORKSPACE_SETTINGS.values())
    return setting_values[0]


def _get_document_key(document: workspace.Document):
    if WORKSPACE_SETTINGS:
        document_workspace = pathlib.Path(document.path)
        workspaces = {s["workspaceFS"] for s in WORKSPACE_SETTINGS.values()}

        # Find workspace settings for the given file.
        while document_workspace != document_workspace.parent:
            if str(document_workspace) in workspaces:
                return str(document_workspace)
            document_workspace = document_workspace.parent

    return None


def _get_settings_by_document(document: workspace.Document | None):
    if document is None or document.path is None:
        return list(WORKSPACE_SETTINGS.values())[0]

    key = _get_document_key(document)
    if key is None:
        # This is either a non-workspace file or there is no workspace.
        key = os.fspath(pathlib.Path(document.path).parent)
        return {
            "cwd": key,
            "workspaceFS": key,
            "workspace": uris.from_fs_path(key),
            **_get_global_defaults(),
        }

    return WORKSPACE_SETTINGS[str(key)]


# *****************************************************
# Internal execution APIs.
# *****************************************************
def _run_tool_on_document(
    document: workspace.Document,
    use_stdin: bool = False,
    extra_args: Optional[Sequence[str]] = None,
) -> utils.RunResult | None:
    """Runs tool on the given document.

    if use_stdin is true then contents of the document is passed to the
    tool via stdin.
    """
    if extra_args is None:
        extra_args = []
    if str(document.uri).startswith("vscode-notebook-cell"):
        # TODO: Decide on if you want to skip notebook cells.
        # Skip notebook cells
        return None

    if utils.is_stdlib_file(document.path):
        # TODO: Decide on if you want to skip standard library files.
        # Skip standard library python files.
        return None

    # deep copy here to prevent accidentally updating global settings.
    settings = copy.deepcopy(_get_settings_by_document(document))

    code_workspace = settings["workspaceFS"]
    cwd = settings["cwd"]

    use_path = False
    use_rpc = False
    if settings["path"]:
        # 'path' setting takes priority over everything.
        use_path = True
        argv = settings["path"]
    elif settings["interpreter"] and not utils.is_current_interpreter(
        settings["interpreter"][0]
    ):
        # If there is a different interpreter set use JSON-RPC to the subprocess
        # running under that interpreter.
        argv = [TOOL_MODULE]
        use_rpc = True
    else:
        # if the interpreter is same as the interpreter running this
        # process then run as module.
        argv = [TOOL_MODULE]

    argv += TOOL_ARGS + settings["args"] + extra_args

    if use_stdin:
        # TODO: update these to pass the appropriate arguments to provide document contents # noqa: E501  # pre-existing
        # to tool via stdin.
        # For example, for pylint args for stdin looks like this:
        #     pylint --from-stdin <path>
        # Here `--from-stdin` path is used by pylint to make decisions on the file contents # noqa: E501  # pre-existing
        # that are being processed. Like, applying exclusion rules.
        # It should look like this when you pass it:
        #     argv += ["--from-stdin", document.path]
        # Read up on how your tool handles contents via stdin. If stdin is not supported use # noqa: E501  # pre-existing
        # set use_stdin to False, or provide path, what ever is appropriate for your tool. # noqa: E501  # pre-existing
        argv += []
    else:
        argv += [document.path]

    if use_path:
        # This mode is used when running executables.
        log_to_output(" ".join(argv))
        log_to_output(f"CWD Server: {cwd}")
        result = utils.run_path(
            argv=argv,
            use_stdin=use_stdin,
            cwd=cwd,
            source=document.source.replace("\r\n", "\n"),
        )
        if result.stderr:
            log_to_output(result.stderr)
    elif use_rpc:
        # This mode is used if the interpreter running this server is different from
        # the interpreter used for running this server.
        log_to_output(" ".join(settings["interpreter"] + ["-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")

        result = jsonrpc.run_over_json_rpc(
            workspace=code_workspace,
            interpreter=settings["interpreter"],
            module=TOOL_MODULE,
            argv=argv,
            use_stdin=use_stdin,
            cwd=cwd,
            source=document.source,
        )
        if result.exception:
            log_error(result.exception)
            result = utils.RunResult(result.stdout, result.stderr)
        elif result.stderr:
            log_to_output(result.stderr)
    else:
        # In this mode the tool is run as a module in the same process as the language server. # noqa: E501  # pre-existing
        log_to_output(" ".join([sys.executable, "-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        # This is needed to preserve sys.path, in cases where the tool modifies
        # sys.path and that might not work for this scenario next time around.

        # with utils.substitute_attr(sys, "path", sys.path[:]):
        #     try:
        #         # TODO: `utils.run_module` is equivalent to running `python -m atopile`. # noqa: E501  # pre-existing
        #         # If your tool supports a programmatic API then replace the function below # noqa: E501  # pre-existing
        #         # with code for your tool. You can also use `utils.run_api` helper, which # noqa: E501  # pre-existing
        #         # handles changing working directories, managing io streams, etc.
        #         # Also update `_run_tool` function and `utils.run_module` in `lsp_runner.py`. # noqa: E501  # pre-existing
        #         result = utils.run_module(
        #             module=TOOL_MODULE,
        #             argv=argv,
        #             use_stdin=use_stdin,
        #             cwd=cwd,
        #             source=document.source,
        #         )
        #     except Exception:
        #         log_error(traceback.format_exc(chain=True))
        #         raise

        if result.stderr:
            log_to_output(result.stderr)

    log_to_output(f"{document.uri} :\r\n{result.stdout}")
    return result


def _run_tool(extra_args: Sequence[str]) -> utils.RunResult:
    """Runs tool."""
    # deep copy here to prevent accidentally updating global settings.
    settings = copy.deepcopy(_get_settings_by_document(None))

    code_workspace = settings["workspaceFS"]
    cwd = settings["workspaceFS"]

    use_path = False
    use_rpc = False
    if len(settings["path"]) > 0:
        # 'path' setting takes priority over everything.
        use_path = True
        argv = settings["path"]
    elif len(settings["interpreter"]) > 0 and not utils.is_current_interpreter(
        settings["interpreter"][0]
    ):
        # If there is a different interpreter set use JSON-RPC to the subprocess
        # running under that interpreter.
        argv = [TOOL_MODULE]
        use_rpc = True
    else:
        # if the interpreter is same as the interpreter running this
        # process then run as module.
        argv = [TOOL_MODULE]

    argv += extra_args

    if use_path:
        # This mode is used when running executables.
        log_to_output(" ".join(argv))
        log_to_output(f"CWD Server: {cwd}")
        result = utils.run_path(argv=argv, use_stdin=True, cwd=cwd)
        if result.stderr:
            log_to_output(result.stderr)
    elif use_rpc:
        # This mode is used if the interpreter running this server is different from
        # the interpreter used for running this server.
        log_to_output(" ".join(settings["interpreter"] + ["-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        result = jsonrpc.run_over_json_rpc(
            workspace=code_workspace,
            interpreter=settings["interpreter"],
            module=TOOL_MODULE,
            argv=argv,
            use_stdin=True,
            cwd=cwd,
        )
        if result.exception:
            log_error(result.exception)
            result = utils.RunResult(result.stdout, result.stderr)
        elif result.stderr:
            log_to_output(result.stderr)
    else:
        # In this mode the tool is run as a module in the same process as the language server. # noqa: E501  # pre-existing
        log_to_output(" ".join([sys.executable, "-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        # This is needed to preserve sys.path, in cases where the tool modifies
        # sys.path and that might not work for this scenario next time around.
        with utils.substitute_attr(sys, "path", sys.path[:]):
            try:
                # TODO: `utils.run_module` is equivalent to running `python -m atopile`.
                # If your tool supports a programmatic API then replace the function below # noqa: E501  # pre-existing
                # with code for your tool. You can also use `utils.run_api` helper, which # noqa: E501  # pre-existing
                # handles changing working directories, managing io streams, etc.
                # Also update `_run_tool_on_document` function and `utils.run_module` in `lsp_runner.py`. # noqa: E501  # pre-existing
                result = utils.run_module(
                    module=TOOL_MODULE, argv=argv, use_stdin=True, cwd=cwd
                )
            except Exception:
                log_error(traceback.format_exc(chain=True))
                raise
        if result.stderr:
            log_to_output(result.stderr)

    log_to_output(f"\r\n{result.stdout}\r\n")
    return result


# *****************************************************
# Logging and notification.
# *****************************************************
def log_to_output(
    message: str, msg_type: lsp.MessageType = lsp.MessageType.Log
) -> None:
    LSP_SERVER.show_message_log("LSP: " + message, msg_type)


def log_error(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Error)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["onError", "onWarning", "always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Error)


def log_warning(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Warning)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["onWarning", "always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Warning)


def log_always(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Info)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Info)


# *****************************************************
# Start the server.
# *****************************************************
if __name__ == "__main__":
    LSP_SERVER.start_io()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math

import faebryk.library._F as F

logger = logging.getLogger(__name__)


class FilterElectricalRC(F.Filter):
    """
    Basic Electrical RC filter
    """

    in_: F.ElectricSignal
    out: F.ElectricSignal
    resistor: F.Resistor
    capacitor: F.Capacitor

    def __preinit__(self):
        self.response.alias_is(F.Filter.Response.LOWPASS)
        self.order.alias_is(1)

        R = self.resistor.resistance
        C = self.capacitor.capacitance
        fc = self.cutoff_frequency

        # Equations
        C.alias_is(1 / (R * 2 * math.pi * fc))
        R.alias_is(1 / (C * 2 * math.pi * fc))
        fc.alias_is(1 / (2 * math.pi * R * C))

        # Connections
        self.in_.line.connect_via(
            (self.resistor, self.capacitor),
            self.in_.reference.lv,
        )

        self.in_.line.connect_via(self.resistor, self.out.line)

        # Set the max voltage of the capacitor to min 1.5 times the output voltage
        self.capacitor.max_voltage.constrain_ge(self.out.reference.voltage * 1.5)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import unittest

HELLO = "Hello, World!"
BYE = "Bye, World!"


def ex():
    print(BYE)


def sync_stuff():
    print("Done")


def async_stuff():
    # FIXME replace this
    # JLCPCB_DB.get()
    # ComponentQuery().filter_by_category(
    #    "Instrumentation/Meter", "Tester"
    # ).filter_by_lcsc_pn("92738").get()
    print("Done")


def normal():
    sync_stuff()


def exc():
    sync_stuff()
    raise Exception()


def async_normal():
    async_stuff()


def async_exc():
    async_stuff()
    raise Exception()


SWITCH = {
    0: normal,
    1: exc,
    2: async_normal,
    3: async_exc,
}


class TestUtilAtExit(unittest.TestCase):
    @unittest.skip("FIXME")
    def test_configs(self):
        import subprocess

        ps = [
            subprocess.Popen(
                ["python", __file__, str(i)],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
            for i in SWITCH
        ]
        for i, p in enumerate(ps):
            out = p.communicate(timeout=10)
            msg = out[0].decode()
            print(i, "-" * 20)
            print(msg)
            lines = msg.splitlines()
            assert lines[-1] == BYE
            assert lines.count(BYE) == 1


# -----------------------------------------------------------------------------



import logging

import typer

logger = logging.getLogger(__name__)


lsp_app = typer.Typer(rich_markup_mode="rich")


@lsp_app.command()
def start():
    from atopile.lsp import LSP_SERVER

    LSP_SERVER.start_io()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import pytest

import faebryk.library._F as F
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.test.times import Times
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "mif_type",
    [
        GraphInterface,
        ModuleInterface,
        F.Electrical,
        F.ElectricPower,
        F.ElectricLogic,
        F.I2C,
    ],
)
def test_performance_mifs_connect_check(mif_type):
    cnt = 100
    timings = Times(multi_sample_strategy=Times.MultiSampleStrategy.AVG)
    name = mif_type.__name__

    def _construct():
        out = mif_type(), mif_type()
        timings.add(name, "construct")
        return out

    instances = times(cnt, _construct)

    for inst1, inst2 in instances:
        inst1.connect(inst2)
        timings.add(name, "connect")

    for inst1, inst2 in instances:
        assert inst1.is_connected_to(inst2)
        timings.add(name, "is_connected")

    logger.info(f"\n{timings}")


@pytest.mark.parametrize(
    "mif_type",
    [
        GraphInterface,
        ModuleInterface,
        F.Electrical,
        F.ElectricPower,
        F.ElectricLogic,
        F.I2C,
    ],
)
def test_performance_mifs_connect_hull(mif_type):
    cnt = 30
    timings = Times()
    name = mif_type.__name__

    def _construct():
        out = mif_type()
        timings.add(name, "construct")
        return out

    instances = times(cnt, _construct)

    for other in instances[1:]:
        instances[0].connect(other)
        timings.add(name, "connect")

    assert instances[0].is_connected_to(instances[-1])
    timings.add(name, "is_connected")

    if issubclass(mif_type, ModuleInterface):
        list(instances[0].get_connected())
    else:
        instances[0].edges
    timings.add(name, "get_connected")

    assert instances[0].is_connected_to(instances[-1])
    timings.add(name, "is_connected cached")

    logger.info(f"\n{timings}")


@pytest.mark.parametrize(
    "module_type",
    [
        F.USB2514B,
        F.RP2040,
    ],
)
def test_performance_mifs_bus_params(module_type):
    timings = Times()
    name = module_type.__name__

    app = module_type()
    timings.add(name, "construct")

    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())
    timings.add(name, "resolve")

    logger.info(f"\n{timings}")


@pytest.mark.slow
def test_performance_mifs_no_connect():
    CNT = 30
    timings = Times(multi_sample_strategy=Times.MultiSampleStrategy.ALL)

    app = F.RP2040_ReferenceDesign()
    timings.add("construct")

    for i in range(CNT):
        list(app.rp2040.power_core.get_connected())
        timings.add("get_connected")

    logger.info(f"\n{timings}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import os
import sys
from pathlib import Path

# @kicad10
KICAD_VERSION = "9.0"

# footprint library paths
# ref: https://docs.kicad.org/8.0/en/kicad/kicad.html#config-file-location
match sys.platform:
    case "win32":
        appdata = os.getenv("APPDATA")
        if appdata is not None:
            GLOBAL_FP_LIB_PATH = (
                Path(appdata) / "kicad" / KICAD_VERSION / "fp-lib-table"
            )
        else:
            raise EnvironmentError("APPDATA environment variable not set")
        # TODO: check on a windows machine
        GLOBAL_FP_DIR_PATH = Path(appdata) / "kicad" / KICAD_VERSION / "footprints"
    case "linux":
        GLOBAL_FP_LIB_PATH = (
            Path("~/.config/kicad").expanduser() / KICAD_VERSION / "fp-lib-table"
        )
        GLOBAL_FP_DIR_PATH = Path("/usr/share/kicad/footprints")
    case "darwin":
        GLOBAL_FP_LIB_PATH = (
            Path("~/Library/Preferences/kicad").expanduser()
            / KICAD_VERSION
            / "fp-lib-table"
        )
        GLOBAL_FP_DIR_PATH = Path(
            "/Applications/KiCad/KiCad.app/Contents/SharedSupport/footprints"
        )
    case _:
        raise EnvironmentError(f"Unsupported platform: {sys.platform}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions

logger = logging.getLogger(__name__)


@dataclass
class FBRKNetlist:
    @dataclass(frozen=True)
    class Component:
        name: str
        value: str
        properties: dict[str, str]

        def __hash__(self) -> int:
            return hash((self.name, self.value, self.properties["footprint"]))

    @dataclass(frozen=True)
    class Net:
        @dataclass(frozen=True)
        class Vertex:
            component: "FBRKNetlist.Component"
            pin: str

        properties: dict[str, str]
        vertices: list[Vertex]

    nets: list[Net]
    comps: list[Component]


def make_fbrk_netlist_from_graph(G: Graph) -> FBRKNetlist:
    from faebryk.exporters.netlist.graph import can_represent_kicad_footprint

    nets = GraphFunctions(G).nodes_of_type(F.Net)

    fbrk_nets = [
        FBRKNetlist.Net(
            properties={"name": net.get_trait(F.has_overriden_name).get_name()},
            vertices=sorted(
                [
                    FBRKNetlist.Net.Vertex(
                        component=t.get_kicad_obj(),
                        pin=t.get_pin_name(mif),
                    )
                    for mif, fp in net.get_connected_pads().items()
                    if (t := fp.get_trait(can_represent_kicad_footprint)) is not None
                ],
                key=lambda v: (v.component.name, v.pin),
            ),
        )
        for net in nets
    ]

    comps = {
        t.get_footprint().get_trait(can_represent_kicad_footprint).get_kicad_obj()
        for _, t in GraphFunctions(G).nodes_with_trait(F.has_footprint)
    }

    not_found = [
        vertex.component
        for net in fbrk_nets
        for vertex in net.vertices
        if vertex.component.name not in {c.name for c in comps}
    ]
    assert not not_found, f"Could not match: {not_found}"

    return FBRKNetlist(nets=fbrk_nets, comps=list(comps))



import os
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

from atopile import errors
from atopile.cli.build import _init_python_app
from atopile.config import config

PROJECT_DIR = Path("test/common/resources/test-project")


@pytest.fixture()
def from_project_dir(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    tmp_project_dir = tmp_path / "test-project"
    shutil.copytree(PROJECT_DIR, tmp_project_dir)
    monkeypatch.chdir(tmp_project_dir)

    yield

    shutil.rmtree(tmp_project_dir)
    monkeypatch.undo()


@pytest.mark.parametrize(
    "build_name,expected_error",
    [
        ("unconstructable", errors.UserPythonConstructionError),
        ("unimportable", errors.UserPythonModuleError),
    ],
)
@pytest.mark.usefixtures("from_project_dir")
def test_build_errors(build_name: str, expected_error):
    config.project_dir = Path.cwd()
    config.selected_builds = [build_name]

    with pytest.raises(expected_error) as exc_info, next(config.builds):
        _init_python_app()

    assert exc_info.value.__cause__ is not None
    assert isinstance(exc_info.value.__cause__, ValueError)
    assert exc_info.value.__cause__.args == (build_name,)


@pytest.mark.parametrize("build_name", ["unconstructable", "unimportable"])
@pytest.mark.usefixtures("from_project_dir")
def test_build_error_logging(build_name: str):
    process = subprocess.run(
        [sys.executable, "-m", "atopile", "build", "-b", build_name],
        capture_output=True,
        text=True,
        env={**os.environ, "NONINTERACTIVE": "1"},
    )

    # single error
    assert process.stderr.count("ERROR") == 1

    # single traceback
    assert process.stderr.count("❱") == 1
    assert process.stderr.count("Traceback (most recent call last)") == 1
    assert "another exception occurred" not in process.stderr
    assert "direct cause of the following exception" not in process.stderr

    # including the test exception
    assert f'raise ValueError("{build_name}")' in process.stderr

    # exiting cleanly
    expected_ending = (
        "Unfortunately errors ^^^ stopped the build. If you need a"
        " hand jump on Discord! https://discord.gg/JY62WubxsP 👋"
    )
    actual_ending = process.stdout.strip().replace("\n", "")
    assert actual_ending.endswith(expected_ending)

    # exception groups are unwrapped
    assert "ExceptionGroup" not in process.stderr

    # with a non-zero exit code
    assert process.returncode == 1



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class PANASONIC_AQY212EHAX(Module):
    """
    PhotoMOS GE 1 Form A(SPST-NO) 1.25V 60V 850mΩ 550mA SOP-4-2.54mm Solid State Relays
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    led: F.LED
    switch = L.f_field(F.Switch(F.Electrical))()

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    explicit_part = L.f_field(F.has_explicit_part.by_supplier)("C29276")
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.led.anode: ["A"],
                self.led.cathode: ["K"],
                self.switch.unnamed[0]: ["S"],
                self.switch.unnamed[1]: ["S1"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.led.del_trait(F.is_pickable)
        self.switch.del_trait(F.is_pickable)

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.led.max_current.constrain_subset(L.Range(1.2 * P.mA, 3.0 * P.mA))
        self.led.reverse_working_voltage.constrain_subset(L.Range(0 * P.V, 5.0 * P.V))
        self.led.forward_voltage.constrain_subset(L.Range(1.25 * P.V, 2.0 * P.V))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class JTAG(ModuleInterface):
    dbgrq: F.ElectricLogic
    tdo: F.ElectricLogic
    tdi: F.ElectricLogic
    tms: F.ElectricLogic
    tck: F.ElectricLogic
    n_trst: F.ElectricLogic
    n_reset: F.ElectricLogic
    vtref: F.ElectricPower

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from hashlib import sha256

from faebryk.core.module import Module


class _FileManuallyModified(Exception): ...


class is_auto_generated(Module.TraitT.decless()):
    CHECKSUM_PLACEHOLDER = "{IS_AUTO_GENERATED_CHECKSUM}"

    def __init__(
        self,
        source: str | None = None,
        system: str | None = None,
        date: str | None = None,
        checksum: str | None = None,
    ) -> None:
        super().__init__()
        self._source = source
        self._system = system
        self._date = date
        self._checksum = checksum
        """
        checksum is the sha256 hash of the file where the checksum string is:
          '{CHECKSUM}'
        """

    @staticmethod
    def _checksum_algo(to_hash: str) -> str:
        return sha256(to_hash.encode("utf-8")).hexdigest()

    @staticmethod
    def verify(stated_checksum: str, file_contents: str):
        actual_checksum = is_auto_generated._checksum_algo(
            file_contents.replace(
                stated_checksum, is_auto_generated.CHECKSUM_PLACEHOLDER
            )
        )

        if actual_checksum != stated_checksum:
            raise _FileManuallyModified(
                f"Checksum mismatch {actual_checksum} != {stated_checksum}"
            )

    @staticmethod
    def set_checksum(file_contents: str) -> str:
        actual = is_auto_generated._checksum_algo(file_contents)
        return file_contents.replace(is_auto_generated.CHECKSUM_PLACEHOLDER, actual)



"""
We collect anonymous telemetry data to help improve atopile.
To opt out, add `telemetry: false` to your project's ~/atopile/telemetry.yaml file.

What we collect:
- Hashed user id so we know how many unique users we have
- Hashed project id
- Error logs
- How long the build took
- ato version
- Git has of current commit
"""

import hashlib
import importlib.metadata
import logging
import time
from contextlib import contextmanager
from typing import Optional

import requests
from attrs import asdict, define, field
from posthog import Posthog
from ruamel.yaml import YAML

from faebryk.libs.paths import get_config_dir
from faebryk.libs.util import cast_assert

log = logging.getLogger(__name__)

# Public API key, as it'd be embedded in a frontend
posthog = Posthog(
    api_key="phc_IIl9Bip0fvyIzQFaOAubMYYM2aNZcn26Y784HcTeMVt",
    host="https://us.i.posthog.com",
    sync_mode=True,
)


@define
class TelemetryData:
    project_id: Optional[str]
    user_id: Optional[str]
    git_hash: Optional[str]
    subcommand: Optional[str]
    time: Optional[float] = None
    assertions: list = field(factory=list)
    eqns_vars: int = 0
    assertions_checked: int = 0
    ato_error: int = 0
    crash: int = 0


telemetry_data: TelemetryData | None = None
start_time: Optional[float] = None


def setup_telemetry_data(command: str):
    global telemetry_data
    _start_timer()
    telemetry_data = TelemetryData(
        project_id=get_project_id(),
        user_id=get_user_id(),
        git_hash=get_current_git_hash(),
        subcommand=command,
    )


def log_telemetry():
    try:
        # Check if telemetry is enabled
        if not load_telemetry_setting():
            log.log(0, "Telemetry is disabled. Skipping telemetry logging.")
            posthog.disabled = True
            return

        telemetry_data.time = _end_timer()
        telemetry_dict = asdict(telemetry_data)
        log.log(0, "Logging telemetry data %s", telemetry_dict)
        requests.post(
            "https://log-telemetry-atsuhzfd5a-uc.a.run.app",
            json=telemetry_dict,
            timeout=0.1,
        ).raise_for_status()
    except requests.exceptions.Timeout:
        # We specifically ignore timeouts here because we expect
        # them to happen most of the time.
        # It's not actually a problem because we don't need to know
        # if the telemetry was logged and we don't want to slow atopile
        # down for it
        pass
    except Exception as e:
        log.log(0, "Failed to log telemetry data: %s", e)


def load_telemetry_setting() -> bool:
    # Use platformdirs to find the appropriate config directory
    atopile_config_dir = get_config_dir()
    atopile_yaml = atopile_config_dir / "telemetry.yaml"

    if not atopile_yaml.exists():
        atopile_config_dir.mkdir(parents=True, exist_ok=True)  # Use the new path
        with atopile_yaml.open("w", encoding="utf-8") as f:
            yaml = YAML()
            yaml.dump({"telemetry": True}, f)
        return True

    with atopile_yaml.open(encoding="utf-8") as f:
        yaml = YAML()
        config = yaml.load(f)
        return config.get("telemetry", True)


def _start_timer():
    global start_time  # Declare start_time as global
    start_time = time.time()


def _end_timer():
    try:
        if start_time is None:
            log.log(0, "Timer was not started.")
            return -1
        end_time = time.time()
        execution_time = end_time - start_time
        log.log(0, "Execution time: %s", execution_time)
    except Exception as ex:
        log.log(0, "Failed to get execution time: %s", ex)
        return None
    return execution_time


def get_user_id() -> str:
    """Generate a unique user ID from the git email."""
    try:
        import git

        try:
            repo = git.Repo(search_parent_directories=True)
            config_reader = repo.config_reader()
            return cast_assert(str, config_reader.get_value("user", "email", "unknown"))
        except (git.InvalidGitRepositoryError, git.NoSuchPathError, ValueError):
            return "unknown"
    except ImportError:
        return "unknown"


def get_current_git_hash() -> Optional[str]:
    """Get the current git commit hash."""
    try:
        import git

        try:
            repo = git.Repo(search_parent_directories=True)
            return repo.head.commit.hexsha
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            return None
    except ImportError:
        return None


def commonise_project_url(git_url: str) -> str:
    """
    Commonize the remote which could be in either of these forms:
        - https://github.com/atopile/atopile.git
        - git@github.com:atopile/atopile.git
    ... to "github.com/atopile/atopile"
    """

    if git_url.startswith("https://"):
        git_url = git_url[8:]
    elif git_url.startswith("git@"):
        git_url = git_url[4:]
        git_url = "/".join(git_url.split(":", 1))

    if git_url.endswith(".git"):
        git_url = git_url[:-4]

    return git_url


def get_project_id() -> Optional[str]:
    """Get the hashed project ID from the git URL of the project, if not available, return 'none'."""  # noqa: E501  # pre-existing
    try:
        import git

        try:
            repo = git.Repo(search_parent_directories=True)
            if not repo.remotes:
                return None
            git_url = repo.remotes.origin.url
            if not git_url:
                return None
        except (git.InvalidGitRepositoryError, git.NoSuchPathError, AttributeError):
            return None
    except ImportError:
        # no git executable
        return None

    project_url = commonise_project_url(git_url)

    log.log(0, "Project URL: %s", project_url)

    # Hash the project ID to de-identify it
    return hashlib.sha256(project_url.encode()).hexdigest()


@contextmanager
def log_to_posthog(event: str, properties: dict | None = None):
    start_time = time.time()

    def _make_properties():
        return {
            "duration": time.time() - start_time,
            "project_id": get_project_id(),
            "project_git_hash": get_current_git_hash(),
            "atopile_version": importlib.metadata.version("atopile"),
            **(properties or {}),
        }

    try:
        yield
    except Exception as e:
        posthog.capture_exception(e, get_user_id(), _make_properties())
        raise

    try:
        posthog.capture(
            distinct_id=get_user_id(),
            event=event,
            properties=_make_properties(),
        )
    except Exception as e:
        log.debug("Failed to log telemetry data: %s", e, exc_info=e)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass, field
from typing import Optional

from faebryk.libs.kicad.fileformats_latest import (
    C_arc,
    C_circle,
    C_footprint,
    C_fp_text,
    C_kicad_footprint_file,
    C_line,
    C_rect,
    C_stroke,
    C_xy,
    E_fill,
    gen_uuid,
)
from faebryk.libs.sexp.dataclass_sexp import SEXP_File, sexp_field
from faebryk.libs.util import not_none

logger = logging.getLogger(__name__)


@dataclass
class C_line_v5:
    start: C_xy
    end: C_xy
    layer: str
    width: float

    def convert_to_new(self) -> C_line:
        return C_line(
            start=self.start,
            end=self.end,
            uuid=gen_uuid(),
            layer=self.layer,
            stroke=C_stroke(
                width=self.width,
                type=C_stroke.E_type.solid,
            ),
        )


@dataclass(kw_only=True)
class C_circle_v5:
    center: C_xy
    end: C_xy
    width: float
    fill: E_fill = field(default=E_fill.no)
    layer: str

    def convert_to_new(self) -> C_circle:
        return C_circle(
            center=self.center,
            end=self.end,
            uuid=gen_uuid(),
            stroke=C_stroke(
                width=self.width,
                type=C_stroke.E_type.solid,
            ),
            fill=self.fill,
            layer=self.layer,
        )


@dataclass
class C_arc_v5:
    start: C_xy
    end: C_xy
    width: float
    layer: str
    angle: float

    def _calculate_midpoint(self) -> tuple[C_xy, C_xy, C_xy]:
        start = self.end
        center = self.start

        mid = start.rotate(center, -self.angle / 2.0)
        end = start.rotate(center, -self.angle)

        return start, mid, end

    def convert_to_new(self) -> C_arc:
        start, mid, end = self._calculate_midpoint()
        return C_arc(
            start=start,
            mid=mid,
            end=end,
            uuid=gen_uuid(),
            stroke=C_stroke(
                width=self.width,
                type=C_stroke.E_type.solid,
            ),
            layer=self.layer,
        )


@dataclass
class C_rect_v5:
    start: C_xy
    end: C_xy
    width: float
    fill: E_fill
    layer: str

    def convert_to_new(self) -> C_rect:
        return C_rect(
            start=self.start,
            end=self.end,
            uuid=gen_uuid(),
            stroke=C_stroke(
                width=self.width,
                type=C_stroke.E_type.solid,
            ),
            fill=self.fill,
            layer=self.layer,
        )


@dataclass
class C_kicad_footprint_file_v5(SEXP_File):
    @dataclass(kw_only=True)
    class C_footprint_in_file(C_footprint):
        @dataclass(kw_only=True)
        class C_model_v5(C_footprint.C_model):
            # V5 offset is called at in some older versions
            # TODO consider implementing sexp_field(alt_name="at") instead or union
            at: Optional[C_footprint.C_model.C_offset] = None
            offset: Optional[C_footprint.C_model.C_offset] = None

            def convert_to_new(self) -> C_footprint.C_model:
                if not self.at and not self.offset:
                    raise ValueError("Either at or offset must be provided")
                return C_footprint.C_model(
                    path=self.path,
                    offset=not_none(self.at or self.offset),
                    scale=self.scale,
                    rotate=self.rotate,
                )

        descr: Optional[str] = None
        tags: Optional[list[str]] = None
        tedit: Optional[str] = None

        fp_lines: list[C_line_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_arcs: list[C_arc_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_circles: list[C_circle_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_rects: list[C_rect_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        model: C_model_v5 | None = None

        def convert_to_new(self) -> C_kicad_footprint_file.C_footprint_in_file:
            propertys: dict[str, C_footprint.C_property] = {
                name: C_footprint.C_property(
                    name=name,
                    value=k.text,
                    at=k.at,
                    layer=k.layer,
                    uuid=k.uuid,
                    effects=k.effects,
                )
                for k in self.fp_texts
                if (name := k.type.capitalize()) in ("Reference", "Value")
            } | self.propertys

            texts = [t for t in self.fp_texts if t.type not in ("reference", "value")]
            for t in self.fp_texts:
                if t.type == "reference":
                    texts.append(
                        C_fp_text(
                            type=C_fp_text.E_type.user,
                            text=t.text.replace("REF**", "${REFERENCE}"),
                            at=t.at,
                            layer=t.layer,
                            uuid=t.uuid,
                            effects=t.effects,
                        )
                    )

            return C_kicad_footprint_file.C_footprint_in_file(
                fp_lines=[line.convert_to_new() for line in self.fp_lines],
                fp_arcs=[arc.convert_to_new() for arc in self.fp_arcs],
                fp_circles=[circle.convert_to_new() for circle in self.fp_circles],
                fp_rects=[rect.convert_to_new() for rect in self.fp_rects],
                # fp-file
                tedit=self.tedit,
                descr=self.descr or "",
                tags=self.tags,
                generator="faebryk",
                generator_version="v5",
                # fp
                name=self.name,
                layer=self.layer,
                propertys=propertys,
                attr=self.attr,
                fp_texts=texts,
                fp_poly=self.fp_poly,
                pads=self.pads,
                models=[self.model.convert_to_new()] if self.model else [],
            )

    module: C_footprint_in_file

    def convert_to_new(self) -> C_kicad_footprint_file:
        return C_kicad_footprint_file(footprint=self.module.convert_to_new())



# pylint: disable=logging-fstring-interpolation

"""
This CLI command provides the `ato install` command to:
- install dependencies
- download JLCPCB footprints
"""

import logging
from pathlib import Path
from typing import Annotated

import questionary
import ruamel.yaml
import typer

from atopile import errors
from atopile.config import DependencySpec, config
from atopile.telemetry import log_to_posthog
from faebryk.libs.backend.packages import api
from faebryk.libs.project.dependencies import ProjectDependencies, ProjectDependency
from faebryk.libs.util import md_list

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
yaml = ruamel.yaml.YAML()

dependencies_app = typer.Typer(rich_markup_mode="rich")


@log_to_posthog("cli:install_end")
def install(
    to_install: Annotated[str | None, typer.Argument()] = None,
    jlcpcb: Annotated[
        bool, typer.Option("--jlcpcb", "-j", help="JLCPCB component ID", hidden=True)
    ] = False,
    vendor: Annotated[
        bool,
        typer.Option(
            "--vendor",
            help="Copy the contents of this dependency into the repo",
        ),
    ] = False,
    local: Annotated[
        Path | None,
        typer.Option(
            "--local",
            "-l",
            help="Install from local path",
            hidden=True,  # deprecated
        ),
    ] = None,
    upgrade: Annotated[
        bool, typer.Option("--upgrade", "-u", help="Upgrade dependencies")
    ] = False,
    path: Annotated[Path | None, typer.Argument()] = None,
):
    """
    Install atopile packages or components from jlcpcb.com/parts.
    Deprecated: Use `ato sync` or `ato add` instead
    """
    if jlcpcb:
        raise errors.UserBadParameterError(
            "--jlcpcb flag has been replaced by `ato create component`"
        )

    if vendor:
        raise errors.UserBadParameterError(
            "--vendor flag has been removed."
            # TODO propose alternative
        )

    if local:
        raise errors.UserBadParameterError(
            "--local flag has been removed. Use `ato add file://` instead",
            extra={"markdown": True},
        )

    if not to_install:
        if upgrade:
            raise errors.UserBadParameterError(
                "Upgrade flag is only supported when adding a package."
            )
        return sync(path=path)
    else:
        return add([to_install], upgrade=upgrade, path=path)


@log_to_posthog("cli:sync_end")
def sync(
    # TODO: only relevant when supporting version specs
    # upgrade: Annotated[
    #    bool,
    #    typer.Option(
    #        "--upgrade", "-U", help="Allow package upgrades, ignoring pinned versions"
    #    ),
    # ] = False,
    path: Annotated[
        Path | None, typer.Option("--project-path", "-C", help="Path to the project")
    ] = None,
):
    # TODO don't like uv's help string
    """
    Update the project's environment
    """

    config.apply_options(None)
    if path:
        config.project_dir = path

    try:
        ProjectDependencies(install_missing=True, clean_unmanaged_dirs=True)
    except (
        api.Errors.PackageNotFoundError,
        api.Errors.ReleaseNotFoundError,
        api.Errors.InvalidPackageIdentifierError,
    ) as e:
        raise errors.UserException(f"Error syncing dependencies: {e}") from e

    logger.info("[green]Done syncing![/] :call_me_hand:", extra={"markup": True})


@log_to_posthog("cli:add_end")
def add(
    package: Annotated[
        list[str],
        typer.Argument(
            help="Package identifier of dependency to add.\n\n"
            "* from packages.atopile.io: <package-name>\\[@<version>]\n\n"
            "* from local project: file://<path-to-local-project-dir>/<package-name>\n\n"
            "* from git: git://<git-repo-url>\\[@<branch/tag/commit>]\n",
        ),
    ],
    upgrade: Annotated[
        bool,
        typer.Option(
            "--upgrade", "-U", help="Allow package upgrades, ignoring pinned versions"
        ),
    ] = False,
    path: Annotated[
        Path | None, typer.Option("--project-path", "-C", help="Path to the project")
    ] = None,
):
    """
    Add dependencies to the project
    """

    if not package:
        raise errors.UserException("No package identifier provided")

    config.apply_options(None)
    if path:
        config.project_dir = path

    deps = ProjectDependencies(install_missing=True, clean_unmanaged_dirs=True)
    try:
        deps.add_dependencies(
            *[DependencySpec.from_str(p) for p in package], upgrade=upgrade
        )
    except (
        api.Errors.PackageNotFoundError,
        api.Errors.ReleaseNotFoundError,
        api.Errors.InvalidPackageIdentifierError,
    ) as e:
        raise errors.UserException(f"Error adding dependencies: {e}") from e

    logger.info(
        "[green]Done adding dependencies![/] :call_me_hand:", extra={"markup": True}
    )


@log_to_posthog("cli:remove_end")
def remove(
    package: Annotated[list[str], typer.Argument(help="Name of package to remove")],
    path: Annotated[
        Path | None, typer.Option("--project-path", "-C", help="Path to the project")
    ] = None,
):
    """
    Remove dependencies from the project
    """

    config.apply_options(None)
    if path:
        config.project_dir = path

    deps = ProjectDependencies()
    deps.remove_dependencies(*package)

    logger.info(
        "[green]Done removing dependencies![/] :call_me_hand:", extra={"markup": True}
    )


@log_to_posthog("cli:list_end")
def list():
    """
    List all dependencies in the project
    """
    deps = ProjectDependencies()
    # TODO bug, if A -> B, B deps
    # will not see that B is under root, because of the way DAG checks roots
    import rich.markdown

    rich.print(
        rich.markdown.Markdown(
            md_list(
                deps.dag.to_tree(extra_roots=deps.direct_deps),
                recursive=True,
                mapper=lambda x: x.spec.identifier
                if isinstance(x, ProjectDependency)
                else x,
            )
        )
    )


dependencies_app.command()(add)
dependencies_app.command()(remove)
dependencies_app.command()(sync)
dependencies_app.command()(list)

# --------------------------------------------------------------------------------------


def check_missing_deps_or_offer_to_install():
    deps = ProjectDependencies()
    if deps.not_installed_dependencies:
        interactive = config.interactive
        logger.warning("It appears some dependencies are missing.")

        if (
            interactive
            and questionary.confirm("Install missing dependencies now?").unsafe_ask()
        ):
            # Install project dependencies, without upgrading
            deps.install_missing_dependencies()
        else:
            logger.warning(
                "Run `ato sync` to install them.",
                extra={"markdown": True},
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.libs.library import L


class CD4011(F.Logic74xx):
    def __init__(self):
        super().__init__(
            [lambda: F.ElectricLogicGates.NAND(input_cnt=2) for _ in range(4)]
        )

    simple_value_representation = L.f_field(F.has_simple_value_representation_defined)(
        "cd4011"
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.module import Module


class has_designator(Module.TraitT.decless()):
    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value

    def get_designator(self) -> str:
        return self.value



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass

import faebryk.library._F as F
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.geometry.basic import Geometry

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True)
class LayoutExtrude(Layout):
    """
    Extrude nodes in a direction from a base position.

    :param vector: x (spacing), y (spacing), r (rotation of the node).
    :param base: The base position to extrude from.
    x (start), y (start), r (direction of extrusion), pcb layer.
    :param dynamic_rotation: If True, the rotation will be multiplied by the index
    of the node, resulting in each consecutive node being rotated more.
    """

    vector: tuple[float, float] | tuple[float, float, float]
    base: F.has_pcb_position.Point = F.has_pcb_position.Point(
        (0, 0, 0, F.has_pcb_position.layer_type.NONE)
    )
    dynamic_rotation: bool = False
    reverse_order: bool = False

    def apply(self, *node: Node):
        """
        Tip: Make sure at least one parent of node has an absolute position defined
        """

        # Remove nodes that have a position defined
        node = tuple(n for n in node if not n.has_trait(F.has_pcb_position))

        vector = self.vector if len(self.vector) == 3 else (*self.vector, 0)

        for i, n in enumerate(
            sorted(
                node,
                key=lambda n: n.get_trait(F.has_designator).get_designator()
                if n.has_trait(F.has_designator)
                else n.get_full_name(),
                reverse=self.reverse_order,
            )
        ):
            vec_i = (
                vector[0] * i,
                vector[1] * i,
                (vector[2] * (i if self.dynamic_rotation else 1)) % 360,
                F.has_pcb_position.layer_type.NONE,
            )
            pos = Geometry.abs_pos(self.base, vec_i)

            n.add(F.has_pcb_position_defined_relative_to_parent(pos))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.app.checks import check_design
from faebryk.libs.exceptions import UserDesignCheckException

logger = logging.getLogger(__name__)


def test_requires_external_usage():
    class App(Module):
        class Outer(Module):
            class Inner(Module):
                b: F.Electrical

            a: F.Electrical
            inner: Inner

            def __preinit__(self):
                self.a.add(F.requires_external_usage())

        outer1: Outer
        outer2: Outer

    app = App()

    # no connections
    with pytest.raises((ExceptionGroup, UserDesignCheckException)) as e:
        check_design(
            app.get_graph(), stage=F.implements_design_check.CheckStage.POST_DESIGN
        )
    if isinstance(e.value, ExceptionGroup):
        assert e.group_contains(
            UserDesignCheckException,
            match="Nodes requiring external usage but not used externally",
        )

    # internal connection
    app.outer1.a.connect(app.outer1.inner.b)
    with pytest.raises((ExceptionGroup, UserDesignCheckException)) as e:
        check_design(
            app.get_graph(), stage=F.implements_design_check.CheckStage.POST_DESIGN
        )
    if isinstance(e.value, ExceptionGroup):
        assert e.group_contains(
            UserDesignCheckException,
            match="Nodes requiring external usage but not used externally",
        )

    # path to external
    app.outer1.inner.b.connect(app.outer2.a)
    with pytest.raises((ExceptionGroup, UserDesignCheckException)) as e:
        check_design(
            app.get_graph(), stage=F.implements_design_check.CheckStage.POST_DESIGN
        )
    if isinstance(e.value, ExceptionGroup):
        assert e.group_contains(
            UserDesignCheckException,
            match="Nodes requiring external usage but not used externally",
        )

    # direct external connection
    app.outer1.a.connect(app.outer2.inner.b)
    check_design(
        app.get_graph(), stage=F.implements_design_check.CheckStage.POST_DESIGN
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class PFET(F.MOSFET):
    def __preinit__(self) -> None:
        self.channel_type.alias_is(F.MOSFET.ChannelType.P_CHANNEL)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

from dataclasses_json import (
    CatchAll,
    Undefined,
    dataclass_json,
)
from dataclasses_json import (
    config as dataclasses_json_config,
)
from easyeda2kicad.easyeda.easyeda_api import EasyedaApi
from easyeda2kicad.easyeda.easyeda_importer import (
    EasyedaFootprintImporter,
    EasyedaSymbolImporter,
)
from easyeda2kicad.easyeda.parameters_easyeda import Ee3dModel, EeSymbol, ee_footprint
from easyeda2kicad.kicad.export_kicad_footprint import ExporterFootprintKicad
from easyeda2kicad.kicad.export_kicad_symbol import ExporterSymbolKicad, KicadVersion
from more_itertools import first

import faebryk.library._F as F
from atopile.config import config as Gcfg
from faebryk.core.module import Module
from faebryk.libs.kicad.fileformats_common import compare_without_uuid
from faebryk.libs.kicad.fileformats_latest import C_kicad_footprint_file
from faebryk.libs.kicad.fileformats_sch import C_kicad_sym_file
from faebryk.libs.kicad.fileformats_version import kicad_footprint_file
from faebryk.libs.picker.localpick import PickerOption
from faebryk.libs.picker.picker import (
    PickedPart,
    PickSupplier,
)
from faebryk.libs.util import ConfigFlag, call_with_file_capture, not_none, once

logger = logging.getLogger(__name__)

CRAWL_DATASHEET = ConfigFlag(
    "LCSC_DATASHEET", default=False, descr="Crawl for datasheet on LCSC"
)

WORKAROUND_SMD_3D_MODEL_FIX = True
"""
easyeda2kicad has not figured out 100% yet how to do model translations.
It's unfortunately also not really easy.
A lot of SMD components (especially passives, ICs, etc) seem to be doing just fine with
an x,y translation of 0. However that makes some other SMD components behave even worse.
Since in a typical design most components are passives etc, this workaround can save
a lot of time and manual work.
"""

WORKAROUND_THT_INCH_MM_SWAP_FIX = False
"""
Some THT models seem to be fixed when assuming their translation is mm instead of inch.
Does not really make a lot of sense.
"""


def _decode_easyeda_date(date: str | int | float) -> datetime:
    if isinstance(date, str):
        return datetime.fromisoformat(date)
    return datetime.fromtimestamp(date)


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass
class EasyEDAAPIResponse:
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class PackageDetail:
        updateTime: int
        unknown: CatchAll = None

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class Lcsc:
        number: str
        unknown: CatchAll = None

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class _Symbol:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class Header:
            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class PartInfo:
                manufacturer: str = field(
                    metadata=dataclasses_json_config(field_name="Manufacturer")
                )
                partno: str = field(
                    metadata=dataclasses_json_config(field_name="Manufacturer Part")
                )
                unknown: CatchAll = None

            part_info: PartInfo = field(
                metadata=dataclasses_json_config(field_name="c_para")
            )
            unknown: CatchAll = None

        info: Header = field(metadata=dataclasses_json_config(field_name="head"))
        unknown: CatchAll = None

    packageDetail: PackageDetail
    updated_at: datetime = field(
        metadata=dataclasses_json_config(decoder=_decode_easyeda_date)
    )
    lcsc: Lcsc
    description: str
    symbol: _Symbol = field(metadata=dataclasses_json_config(field_name="dataStr"))

    _atopile_queried_at: float | None = field(
        default=None,
        metadata=dataclasses_json_config(field_name="atopile_queried_at"),
    )
    _atopile_manufacturer: str | None = field(
        default=None,
        metadata=dataclasses_json_config(field_name="atopile_manufacturer"),
    )
    unknown: CatchAll = None

    @property
    def query_time(self) -> datetime:
        return datetime.fromtimestamp(not_none(self._atopile_queried_at))

    @property
    def mfn_pn(self) -> tuple[str, str]:
        pn = self.symbol.info.part_info.partno
        mfr = self._atopile_manufacturer or self.symbol.info.part_info.manufacturer
        if not self._atopile_manufacturer:
            logger.warning(
                f"No manufacturer for ({mfr} {pn}) {self.lcsc.number} found in backend."
            )
        # remove chinese manufacturer name in parentheses
        # "TI(德州仪器)" -> "TI"
        mfr = re.sub(r"\([^)]*\)", "", mfr)
        return (mfr, pn)

    @classmethod
    def from_api_dict(cls, data: dict):
        out: EasyEDAAPIResponse = cls.from_dict(data)  # type: ignore
        out._atopile_queried_at = datetime.timestamp(datetime.now())
        return out

    def serialize(self) -> bytes:
        return self.to_json(indent=4).encode("utf-8")  # type: ignore

    def dump(self, path: Path):
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.serialize())

    @classmethod
    def deserialize(cls, data: bytes):
        out: cls = cls.from_json(data.decode("utf-8"))  # type: ignore
        if out._atopile_queried_at is None:
            out._atopile_queried_at = datetime.timestamp(out.updated_at)
        return out

    @classmethod
    def load(cls, path: Path):
        return cls.deserialize(path.read_bytes())

    def raw(self) -> dict:
        return self.to_dict()  # type: ignore


class EasyEDA3DModel:
    def __init__(self, step: bytes, name: str):
        self.step = step
        self.name = name

    def serialize(self) -> bytes:
        return self.step

    def dump(self, path: Path):
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.step)

    @classmethod
    def deserialize(cls, data: bytes, name: str):
        return cls(data, name)

    @classmethod
    def load(cls, path: Path):
        return cls(path.read_bytes(), path.name)


class EasyEDAFootprint:
    cache: dict[Path, "EasyEDAFootprint"] = {}

    def __init__(self, footprint: C_kicad_footprint_file):
        self.footprint = footprint

    def serialize(self) -> bytes:
        return self.footprint.dumps().encode("utf-8")

    @classmethod
    def load(cls, path: Path):
        # assume not disk modifications during run
        if path in cls.cache:
            return cls.cache[path]
        out = cls(kicad_footprint_file(path))
        cls.cache[path] = out
        return out

    @classmethod
    def from_api(cls, footprint: ee_footprint, model_path: str | None):
        exporter = ExporterFootprintKicad(footprint)
        _fix_3d_model_offsets(exporter)
        if model_path is None:
            # allows our ignored type annotation to work below
            assert footprint.model_3d is None

        fp_raw = call_with_file_capture(
            lambda path: exporter.export(str(path), model_path)  # type: ignore
        )[1]
        fp = kicad_footprint_file(fp_raw.decode("utf-8"))
        # workaround: remove wrl ending easyeda likes to add for no reason
        for m in fp.footprint.models:
            if m.path.suffix == ".wrl":
                m.path = m.path.parent
        return cls(fp)

    def dump(self, path: Path):
        type(self).cache[path] = self
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.serialize())

    @property
    def library_name(self):
        return self.footprint.footprint.name

    @property
    def base_name(self):
        return self.library_name.split(":")[-1]

    def compare(self, other: "EasyEDAFootprint"):
        return compare_without_uuid(
            self.footprint,
            other.footprint,
        )


class EasyEDASymbol:
    cache: dict[Path, "EasyEDASymbol"] = {}

    def __init__(self, symbol: C_kicad_sym_file):
        self.symbol = symbol

    @classmethod
    def from_api(cls, symbol: EeSymbol):
        from faebryk.libs.kicad.fileformats_v6 import C_symbol_in_file_v6
        from faebryk.libs.sexp.dataclass_sexp import loads as sexp_loads

        exporter = ExporterSymbolKicad(symbol, KicadVersion.v6)
        # TODO this is weird
        fp_lib_name = symbol.info.lcsc_id
        raw = exporter.export(footprint_lib_name=fp_lib_name)
        sym = sexp_loads(raw, C_symbol_in_file_v6).symbol.convert_to_new()
        sym_file = C_kicad_sym_file(
            C_kicad_sym_file.C_kicad_symbol_lib(
                version=1,
                generator="faebryk",
                symbols={sym.name: sym},
            )
        )
        return cls(sym_file)

    def serialize(self) -> bytes:
        return self.symbol.dumps().encode("utf-8")

    def dump(self, path: Path):
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.serialize())
        type(self).cache[path] = self

    @classmethod
    def load(cls, path: Path):
        # assume not disk modifications during run
        if path in cls.cache:
            return cls.cache[path]
        out = cls(C_kicad_sym_file.loads(path))
        cls.cache[path] = out
        return out

    def compare(self, other: "EasyEDASymbol"):
        return compare_without_uuid(
            self.symbol,
            other.symbol,
        )

    @property
    def kicad_symbol(self):
        return first(self.symbol.kicad_symbol_lib.symbols.values())


class EasyEDAPart:
    def __init__(
        self,
        lcsc_id: str,
        description: str,
        mfn_pn: tuple[str, str],
        footprint: EasyEDAFootprint,
        symbol: EasyEDASymbol,
        model: EasyEDA3DModel | None = None,
        datasheet_url: str | None = None,
    ):
        self.lcsc_id = lcsc_id
        self.description = description
        self.mfn_pn = mfn_pn
        self.footprint = footprint
        self.symbol = symbol
        self.model = model
        self._pre_model: Ee3dModel | None = None
        self.datasheet_url = datasheet_url
        self._pre_datasheet: str | None = None

    @property
    def identifier(self):
        return self.lcsc_id

    @property
    def model_name(self):
        if self.model:
            return self.model.name
        assert self._pre_model is not None
        return self._pre_model.name

    def load_model(self):
        from faebryk.libs.part_lifecycle import PartLifecycle

        lifecycle = PartLifecycle.singleton()
        assert self.model is None
        assert self._pre_model is not None
        if lifecycle.easyeda2kicad.shall_refresh_model(self):
            logger.debug(f"Downloading model for {self.identifier}")
            model = EasyedaApi().get_step_3d_model(uuid=self._pre_model.uuid)
            # might happen sometimes, that even tho it's in the api, it's not available
            if model is None:
                self.model = None
            else:
                self.model = EasyEDA3DModel(model, self._pre_model.name)
        else:
            self.model = lifecycle.easyeda2kicad.load_model(self)

        self._pre_model = None

    @classmethod
    def from_api_response(cls, data: EasyEDAAPIResponse, download_model: bool):
        """
        args:
            download_model: Purely for performance and api overloading
        """
        from faebryk.libs.part_lifecycle import PartLifecycle

        lifecycle = PartLifecycle.singleton()

        easyeda_footprint = EasyedaFootprintImporter(
            easyeda_cp_cad_data=data.raw()
        ).get_footprint()

        easyeda_symbol = EasyedaSymbolImporter(
            easyeda_cp_cad_data=data.raw()
        ).get_symbol()

        easyeda_model = easyeda_footprint.model_3d

        if easyeda_model is not None:
            model_name = easyeda_model.name
            model_path = lifecycle.easyeda2kicad.get_model_path(
                data.lcsc.number, model_name
            )
            kicad_model_path = str(
                Gcfg.project.get_relative_to_kicad_project(model_path)
            )
        else:
            kicad_model_path = None

        part = cls(
            lcsc_id=data.lcsc.number,
            description=data.description,
            mfn_pn=data.mfn_pn,
            footprint=EasyEDAFootprint.from_api(easyeda_footprint, kicad_model_path),
            symbol=EasyEDASymbol.from_api(easyeda_symbol),
        )
        part._pre_model = easyeda_model
        part._pre_datasheet = easyeda_symbol.info.datasheet

        if download_model and part._pre_model is not None:
            part.load_model()

        return part

    def load_datasheet(self):
        if self.datasheet_url:
            return self.datasheet_url

        # TODO use easyeda2kicad api as soon as works again
        # return symbol.info.datasheet

        if not CRAWL_DATASHEET:
            return None

        import re

        import requests

        url = self._pre_datasheet
        if not url:
            return None

        logger.debug(f"Crawling datasheet for {self.identifier}")

        # make requests act like curl
        lcsc_site = requests.get(url, headers={"User-Agent": "curl/7.81.0"})
        lcsc_id = self.lcsc_id
        # find _{partno}.pdf in html
        match = re.search(f'href="(https://[^"]+_{lcsc_id}.pdf)"', lcsc_site.text)
        if match:
            pdfurl = match.group(1)
            logger.debug(f"Found datasheet for {lcsc_id} at {pdfurl}")
            return pdfurl
        else:
            return None


def _fix_3d_model_offsets(ki_footprint: ExporterFootprintKicad):
    if ki_footprint.output.model_3d is None:
        return

    if WORKAROUND_SMD_3D_MODEL_FIX:
        if ki_footprint.input.info.fp_type == "smd":
            ki_footprint.output.model_3d.translation.x = 0
            ki_footprint.output.model_3d.translation.y = 0
    if WORKAROUND_THT_INCH_MM_SWAP_FIX:
        if ki_footprint.input.info.fp_type != "smd":
            ki_footprint.output.model_3d.translation.x *= 2.54
            ki_footprint.output.model_3d.translation.y *= 2.54


class LCSCException(Exception):
    def __init__(self, partno: str, *args: object) -> None:
        self.partno = partno
        super().__init__(*args)

    def __str__(self) -> str:
        return f"{type(self).__name__}: {self.partno} - {self.args}"


class LCSC_NoDataException(LCSCException): ...


class LCSC_PinmapException(LCSCException): ...


@once
def get_raw(lcsc_id: str) -> EasyEDAAPIResponse:
    from faebryk.libs.part_lifecycle import PartLifecycle

    lifecycle = PartLifecycle.singleton()
    if not lifecycle.easyeda_api.shall_refresh(lcsc_id):
        return lifecycle.easyeda_api.load(lcsc_id)

    logger.debug(f"Downloading API data {lcsc_id}")
    api = EasyedaApi()
    cad_data = api.get_cad_data_of_component(lcsc_id=lcsc_id)
    # API returned no data
    if not cad_data:
        raise LCSC_NoDataException(
            lcsc_id, f"Failed to fetch data from EasyEDA API for part {lcsc_id}"
        )
    response = EasyEDAAPIResponse.from_api_dict(cad_data)
    # TODO: this is a hack
    # get manufacturer from backend
    from faebryk.libs.picker.api.api import get_api_client

    if api_part := first(
        get_api_client().fetch_part_by_lcsc(int(lcsc_id.removeprefix("C"))), None
    ):
        response._atopile_manufacturer = api_part.manufacturer_name

    return lifecycle.easyeda_api.ingest(lcsc_id, response)


def download_easyeda_info(lcsc_id: str, get_model: bool = True):
    from faebryk.libs.part_lifecycle import PartLifecycle

    lifecycle = PartLifecycle.singleton()

    data = get_raw(lcsc_id)
    part = EasyEDAPart.from_api_response(data, download_model=False)

    if get_model and not part._pre_model:
        logger.warning(f"No 3D model for '{lcsc_id} ({part.footprint.base_name})'")

    if get_model and lifecycle.easyeda2kicad.shall_refresh_model(part):
        part.load_model()

    return lifecycle.easyeda2kicad.ingest_part(part)


def check_attachable(component: Module):
    if not component.has_trait(F.has_footprint):
        if not component.has_trait(F.can_attach_to_footprint):
            if not component.has_trait(F.has_pin_association_heuristic):
                raise LCSC_PinmapException(
                    "",
                    f"Need either F.can_attach_to_footprint or "
                    "F.has_pin_association_heuristic"
                    f" for {component}",
                )


def attach(
    component: Module, partno: str, get_model: bool = True, check_only: bool = False
):
    from faebryk.libs.part_lifecycle import PartIsNotAutoGenerated, PartLifecycle

    # TODO fix this later on
    # for now, we always get the model
    # related to #1265
    get_model = True

    lifecycle = PartLifecycle.singleton()
    try:
        epart = download_easyeda_info(partno, get_model=get_model)
    except LCSC_NoDataException:
        if component.has_trait(F.has_footprint):
            apart = None
            epart = None
        else:
            raise
    else:
        try:
            apart = lifecycle.library.ingest_part_from_easyeda(epart)
        except PartIsNotAutoGenerated as ex:
            apart = ex.part
            logger.debug(
                f"Part `{ex.part.path}` is not purely auto-generated.Not overwriting."
            )

    # TODO maybe check the symbol matches, even if a footprint is already attached?
    if not component.has_trait(F.has_footprint):
        assert apart is not None
        if not component.has_trait(F.can_attach_to_footprint):
            # TODO make this a trait
            pins = [
                (pin.number.number, pin.name.name)
                for sym in apart.symbol.kicad_symbol_lib.symbols.values()
                for unit in sym.symbols.values()
                for pin in unit.pins
            ]
            try:
                pinmap = component.get_trait(F.has_pin_association_heuristic).get_pins(
                    pins
                )
            except F.has_pin_association_heuristic.PinMatchException as e:
                raise LCSC_PinmapException(partno, f"Failed to get pinmap: {e}") from e

            if check_only:
                return

            component.add(F.can_attach_to_footprint_via_pinmap(pinmap))

            sym = F.Symbol.with_component(component, pinmap)
            # FIXME
            sym.add(
                F.Symbol.has_kicad_symbol(f"{apart.identifier}:{apart.sym_path.name}")
            )

        if check_only:
            return

        # footprint
        fp = F.KicadFootprint.from_path(apart.fp_path, lib_name=apart.path.name)
        component.get_trait(F.can_attach_to_footprint).attach(fp)

    if check_only:
        return

    # model done by kicad (in fp)


class PickSupplierLCSC(PickSupplier):
    supplier_id: str = "lcsc"

    def attach(self, module: Module, part: PickerOption):
        assert isinstance(part.part, PickedPartLCSC)
        attach(component=module, partno=part.part.lcsc_id)

    def __str__(self) -> str:
        return f"{type(self).__name__}()"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other)


@dataclass(frozen=True, kw_only=True)
class PickedPartLCSC(PickedPart):
    @dataclass(frozen=True)
    class Info:
        stock: int
        price: float
        description: str
        basic: bool
        preferred: bool

    info: Info | None = None
    supplier: PickSupplierLCSC = field(default_factory=PickSupplierLCSC)

    @property
    def lcsc_id(self) -> str:
        return self.supplier_partno



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from textwrap import dedent

from faebryk.libs.codegen.atocodegen import AtoCodeGen


def _cleanup(s: str) -> str:
    return dedent(s).removeprefix("\n")


def test_atogen_basic_part():
    cf = AtoCodeGen.ComponentFile("TI_TCA9548APWR")

    cf.add_trait(
        "is_atomic_part",
        manufacturer="TI",
        partnumber="TCA9548APWR",
        footprint="TSSOP-24_L7.8-W4.4-P0.65-LS6.4-BL.kicad_mod",
        symbol="TCA9548APWR.kicad_sym",
        model="TSSOP-24_L7.8-W4.4-H1.0-LS6.4-P0.65.step",
    )
    cf.add_comments(
        "This trait marks this file as auto-generated",
        "If you want to manually change it, remove the trait",
        use_spacer=True,
    )
    cf.add_trait(
        "is_auto_generated",
        system="ato_part",
        source="easyeda:C130026",
        date="2025-05-26T12:53:38.352371+00:00",
        checksum="{CHECKSUM_PLACEHOLDER}",
    )
    ato = cf.dump()

    assert ato == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part
        import is_auto_generated

        component TI_TCA9548APWR:
            trait is_atomic_part<manufacturer="TI", partnumber="TCA9548APWR", footprint="TSSOP-24_L7.8-W4.4-P0.65-LS6.4-BL.kicad_mod", symbol="TCA9548APWR.kicad_sym", model="TSSOP-24_L7.8-W4.4-H1.0-LS6.4-P0.65.step">

            # This trait marks this file as auto-generated
            # If you want to manually change it, remove the trait
            trait is_auto_generated<system="ato_part", source="easyeda:C130026", date="2025-05-26T12:53:38.352371+00:00", checksum="{CHECKSUM_PLACEHOLDER}">
        """  # noqa: E501
    )


def test_atogen_postmod():
    cf = AtoCodeGen.ComponentFile("BLA")
    t = cf.add_trait("is_atomic_part")

    assert cf.dump() == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part

        component BLA:
            trait is_atomic_part
        """
    )

    t.args = {
        "manufacturer": "TI",
        "partnumber": "TCA9548APWR",
    }

    assert cf.dump() == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part

        component BLA:
            trait is_atomic_part<manufacturer="TI", partnumber="TCA9548APWR">
        """
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P


class USB3(ModuleInterface):
    usb3_if: F.USB3_IF

    def __preinit__(self):
        self.usb3_if.gnd_drain.connect(self.usb3_if.usb_if.buspower.lv)
        self.usb3_if.usb_if.buspower.voltage.constrain_subset(
            L.Range(4.75 * P.V, 5.5 * P.V)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import inspect
from pathlib import Path

import pytest

from atopile.front_end import Bob, TypeRef, ap
from faebryk.core.core import Namespace
from faebryk.core.node import Node
from faebryk.core.trait import Trait
from faebryk.libs.library import L
from faebryk.libs.util import not_none

try:
    import faebryk.library._F as F
except ImportError:
    F = None


def test_load_library():
    assert F is not None, "Failed to load library"


@pytest.mark.skipif(F is None, reason="Library not loaded")
@pytest.mark.parametrize("name, module", list(vars(F).items()))
def test_symbol_types(name: str, module):
    # private symbols get a pass
    if name.startswith("_"):
        return

    # skip once wrappers
    # allow once wrappers for type generators
    if getattr(module, "_is_once_wrapper", False):
        return

    # otherwise, only allow Node or Namespace class objects
    assert isinstance(module, type) and issubclass(module, (Node, Namespace))


@pytest.mark.skipif(F is None, reason="Library not loaded")
@pytest.mark.parametrize(
    "name, module",
    [
        (name, module)
        for name, module in vars(F).items()
        if not (
            name.startswith("_")
            or not isinstance(module, type)
            or not issubclass(module, Node)
            or (issubclass(module, Trait))
        )
    ],
)
# @pytest.mark.timeout(60)  # TODO lower
def test_init_args(name: str, module):
    """Make sure we can instantiate all classes without error"""

    # handle post_init_decorator
    init = (
        module.__init__
        if not hasattr(module, "__original_init__")
        else module.__original_init__
    )
    init_signature = inspect.signature(init)
    args = [p for p in init_signature.parameters.values() if p.name != "self"]

    # check if constructor has no non-default args
    if any(
        p.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        and p.default is inspect.Parameter.empty
        for p in args
    ):
        pytest.skip(
            f"Skipped module with init args because we can't instantiate it: {args}"
        )

    try:
        module()
    except L.AbstractclassError:
        pytest.skip("Skipped abstract class")


def _module_addr(file: Path, module_name: TypeRef):
    return f"{file.name}:{module_name}"


_file_and_modules = [
    (file, module_name)
    for file in Path(inspect.getfile(not_none(F))).parent.glob("*.ato")
    for module_name, module in Bob().index_file(file).refs.items()
    if isinstance(module, ap.BlockdefContext)
]


@pytest.mark.skipif(F is None, reason="Library not loaded")
@pytest.mark.parametrize(
    "file, module_name",
    _file_and_modules,
    ids=[_module_addr(file, module_name) for file, module_name in _file_and_modules],
)
def test_instance_library_ato(file: Path, module_name: TypeRef):
    bob = Bob()
    bob.build_file(file, module_name)
